##############################################################################
##
## Unnstall-PSQL
##
## by Actian Corporation (http://www.actian.com/psql)
##
##############################################################################

    <#
    .NOTES
        Copyright (c) Actian Corporation.  All rights reserved.

        Use of this sample source code is subject to the terms of the Actian
        PSQL license agreement under which you licensed this sample source code.
        If you did not accept the terms of the license agreement, you are not
        authorized to use this sample source code. For the terms of the license,
        please see the license agreement between you and Actian Corporation.
        THE SAMPLE SOURCE CODE IS PROVIDED "AS IS", WITH NO WARRANTIES.

    .SYNOPSIS
        Uninstalls PSQL.

    .DESCRIPTION
        Uninstalls any detected PSQL product type from the local system.  If a PSQL
        engine is installed, the DBNAMES.cfg (in the PSQL data folder) is left on
        the system.  If more than one PSQL product type is detected, a warning is
        displayed and only one of the detected products type is removed.  The
        script can be run multiple times to ensure all PSQL product types are
        removed.

    .EXAMPLE
        PS C:\> .\Uninstall-PSQL
            Uninstalls PSQL.

    .INPUTS
        None. You cannot pipe objects to this cmdlet.

    .OUTPUTS
        Creates log file in users $env:temp folder.
    #>

#Requires -Version 5.1        # -- Require PowerShell 5.1 or higher (Nano Server,IoT).
#Requires -PSEdition Core     # -- Require PowerShell Core
#Requires -RunAsAdministrator # -- Require PowerShell running as Administrator.

Set-StrictMode -Version Latest

Import-LocalizedData msg -FileName "PSQL_PowerShell_Script_strings.psd1" -ErrorAction Ignore

#############################
## Variables Required prior to calling SetScriptVariables().
#############################
if ([System.IntPtr]::Size -eq 4) {$script:Is64Bit=$false} else {$script:Is64Bit=$true}
$script:registryPath = "HKLM:\SOFTWARE\Pervasive Software"
$script:culture=Get-Culture
if ( $PSCulture.StartsWith("ja") ) {$script:IsJajp=$true} else {$script:IsJajp=$false}
$script:removing=$true

#############################
## FUNCTIONS
#############################
function Main {
    ##########################
    # Defines the processing order of this script.
    ##########################
    FindProductType
    SetScriptVariables
    BeginLog
    StartPsqlServices
    InvokeECollect
    UnloadPerfMonCounterSets
    #RemovePsqlDatabases
    UnregisterComInterfaces
    StopPsqlServices
    RemovePsqlFirewallRules
    RemovePsqlServices
    if ($Is64Bit) {
        $bin64=$PsqlInstDir64+"bin\"
        RemoveStrFromPath $bin64
    } else {
        $bin32=$PsqlInstDir32+"bin\"
        RemoveStrFromPath $bin32
    }
    RemovePsqlRegistry
    RemovePsqlFiles
    EndLog
}
function FindProductType {
    ##########################
    # Sets the installed product type based on the last instance of a key matching the string
    # "*Pervasive.SQL*" being found under 'hklm:\SOFTWARE\Pervasive Software\Products'.
    #
    # Variables set:
    #   [string] $script:InstalledProductType = client|server|reporting
    #   [object] $global:PsqlInfo = object methods match InstallInfo registry values.
    #                   $PsqlInfo.InstallData
    #                   $PsqlInfo.InstallDir64
    #                   $PsqlInfo.InstallDir
    #                   $PsqlInfo.IsBtrieve
    #                   $PsqlInfo.IsReport
    #                   $PsqlInfo.IsVx
    #                   $PsqlInfo.ProductVersion
    #                   $PsqlInfo.VersionLevel
    #                   $PsqlInfo.InstallData
    #
    # Note: Must be called before logging is started.
    ##########################
    $script:InstalledProductType=""
    if (test-path $registryPath"\Products\") {
        Get-ChildItem $registryPath"\Products\" | ForEach-Object -Process {
            if ($_.Name -Like "*Pervasive.SQL*" ) {
                switch ($_.PSChildName) {
                    'Pervasive.SQL Client' {
                        $global:PsqlInfo=Get-ItemProperty $registryPath"\Products\"$_"\InstallInfo"
                        $script:InstalledProductType="client"
                        break}
                    'Pervasive.SQL NT Server' {
                        $global:PsqlInfo=Get-ItemProperty $registryPath"\Products\"$_"\InstallInfo"
                        if ($PsqlInfo.IsVx -eq 0) {
                            $script:InstalledProductType="server"
                        } else {
                            $script:InstalledProductType="vx"
                        }
                        break}
                    'Pervasive.SQL Workgroup' {
                        $global:PsqlInfo=Get-ItemProperty $registryPath"\Products\"$_"\InstallInfo"
                        if ($PsqlInfo.IsReport -eq 0) {
                            $script:InstalledProductType="workgroup"
                        } else {
                            $script:InstalledProductType="reporting"
                        }
                        break}
                }
            }
        }
    }
}
function SetScriptVariables {
    ##########################
    # Sets script-level variables based on specified product type to be installed. Except
    # for $PsqlInstallCfg, of the variable is set to a simple string then it can be
    # overridden by specifying the name=value pair in the install configuration file
    # specified by $PsqlInstallCfg.
    #
    # Note: Must be called before logging is started.
    ##########################
    #$script:PsqlInstallCfg="install_"+ $InstalledProductType + ".cfg"
    #$script:defaultAclArgs="Everyone","FullControl","ContainerInherit","InheritOnly","Allow"
    $script:scriptName = & {$myInvocation.ScriptName}
    $script:scriptPath = & {Split-Path $myInvocation.ScriptName}
    if ($env:PROCESSOR_ARCHITECTURE.Contains("64")) {$script:Is64Bit=$true} else {$script:Is64Bit=$false}
    [string]$fwpDomainState=netsh advfirewall show domainprofile State
    if ( $fwpDomainState.Contains("ON") ) {$script:fwDomainProfileState="ON"} else {$script:fwDomainProfileState="OFF"}
    [string]$fwpPrivateState=netsh advfirewall show privateprofile State
    if ( $fwpPrivateState.Contains("ON") ) {$script:fwPrivateProfileState="ON"} else {$script:fwPrivateProfileState="OFF"}
    [string]$fwpPublicState=netsh advfirewall show publicprofile State
    if ( $fwpPublicState.Contains("ON") ) {$script:fwPublicProfileState="ON"} else {$script:fwPublicProfileState="OFF"}
    #$script:registryPath = "HKLM:\SOFTWARE\Pervasive Software"
    $script:CmnSvcDispName = "Actian PSQL"
    $script:Elsh=""
    $script:Elshid="0"
    # -- $PsqlInfo properties set by FindProductTye
    $script:PsqlDataDir=$PsqlInfo.InstallData
    $script:PsqlInstDir64=$PsqlInfo.InstallDir64
    $script:PsqlInstDir32=$PsqlInfo.InstallDir
    $script:setupdbCfgInstall=$PsqlDataDir + "w32setupdb.cfg"
    # -- Default values (overwritten based on $InstalledProductType)
    $script:isClient=$False
    $script:isServer=$False
    $script:isVxServer=$False
    $script:isWorkgroup=$False
    $script:isReporting=$False
    $script:uninstallLogFile = $env:TEMP + "\PSQL_PowerShell_Uninstall.log"
    # -- Variables based on command line arugments.
    switch ($InstalledProductType) {
        'client' {
            $script:isClient=$True
            $script:InstallInfoProdType="Pervasive.SQL Client"
            $script:uninstallLogFile = $env:TEMP + "\PSQL_Client_PowerShell_Uninstall.log"
            break}
        'server' {
            $script:isServer=$True
            $script:InstallInfoProdType="Pervasive.SQL NT Server"
            $script:uninstallLogFile = $env:TEMP + "\PSQL_ServerEngine_PowerShell_Uninstall.log"
            break}
        'vx' {
            $script:isVxServer=$True
            $script:InstallInfoProdType="Pervasive.SQL NT Server"
            $script:uninstallLogFile = $env:TEMP + "\PSQL_VxServer_PowerShell_Uninstall.log"
            break}
        'workgroup' {
            $script:isWorkgroup=$True
            $script:InstallInfoProdType="Pervasive.SQL Workgroup"
            $script:uninstallLogFile = $env:TEMP + "\PSQL_WorkgroupEngine_PowerShell_Uninstall.log"
            break}
        'reporting' {
            $script:isReporting=$True
            $script:InstallInfoProdType="Pervasive.SQL Workgroup"
            $script:uninstallLogFile = $env:TEMP + "\PSQL_ReportingEngine_PowerShell_Uninstall.log"
            break}
    }
    # -- COM self-registering file lists.
    $script:ComFileList="acbtr732.ocx","acctr732.ocx","clientrb.dll","cobolschemaexecmsgrb.dll","dbcsipxy.dll","dto.dll","dto2.dll","legacylm.dll","licmgrrb.dll","mkc3.dll","mkderb.dll","pceurop.dll","pctlgrb.dll","poledb.dll","poledber.dll","poledbsy.dll","pscp932.dll","pscp936.dll","pscp949.dll","pscp950.dll","pscrb_en.dll","pscrb_ja.dll","psdom.dll","pseucjp.dll","pssax.dll","pstestct.dll","pstestsp.dll","pstsprb.dll","pvconnts_en.dll","pvconnts_ja.dll","pvinfodg_en.dll","pvinfodg_ja.dll","pvmsgrb.dll","upiapirb.dll","w3cobolschemaexec100.dll","w3csi100.dll","w3csm100.dll","w3csp100.dll","w3dcm100.dll","w3exp010.dll","w3mpm100.dll","w64clientrb.dll","w64cobolschemaexec100.dll","w64cobolschemaexecmsgrb.dll","w64csi.dll","w64csm.dll","w64csp.dll","w64dbcsipxy.dll","w64dcm.dll","w64dto2.dll","w64exp010.dll","w64legacylm.dll","w64licmgrrb.dll","w64mkc3.dll","w64mkderb.dll","w64mpm.dll","w64pceurop.dll","w64pctlgrb.dll","w64pscp932.dll","w64pscp936.dll","w64pscp949.dll","w64pscp950.dll","w64psdom.dll","w64pseucjp.dll","w64pssax.dll","w64pvmsgrb.dll","w64upiapirb.dll"
}
function BeginLog {
    ##########################
    # Create $uninstallLogFile and write basic info to the file.
    ##########################
    "#############################" | Out-File -Append $uninstallLogFile
    $msg.infoStart -f [DateTime]::Parse((Get-Date).ToString()) | Out-File -Append $uninstallLogFile
    $msg.infoScriptName -f $scriptName | Out-File -Append $uninstallLogFile
    $msg.infoScriptPath -f $ScriptPath | Out-File -Append $uninstallLogFile
    $msg.infoCompName -f $env:COMPUTERNAME | Out-File -Append $uninstallLogFile
    $msg.infoCulture -f $culture.Name | Out-File -Append $uninstallLogFile
    $msg.infoProcNum -f [System.Environment]::ProcessorCount | Out-File -Append $uninstallLogFile
    $msg.infoWinArch -f $env:PROCESSOR_ARCHITECTURE | Out-File -Append $uninstallLogFile
    $msg.infoWinPlat -f [System.Environment]::OSVersion.Platform | Out-File -Append $uninstallLogFile
    $msg.infoWinVer -f [System.Environment]::OSVersion.Version | Out-File -Append $uninstallLogFile
    $msg.infoIs64Bit -f $Is64Bit | Out-File -Append $uninstallLogFile 
    $msg.infoCLRVer -f [System.Environment]::Version | Out-File -Append $uninstallLogFile
    $msg.infoUserName -f $env:USERNAME | Out-File -Append $uninstallLogFile
    $msg.infoUserDomain -f $env:USERDOMAIN | Out-File -Append $uninstallLogFile
    $msg.infoPSVer -f $psVersionTable.PSVersion | Out-File -Append $uninstallLogFile
    $msg.infoPSCLRVer -f $psVersionTable.CLRVersion | Out-File -Append $uninstallLogFile
    $msg.infoFWDomainPS -f $fwDomainProfileState | Out-File -Append $uninstallLogFile
    $msg.infoFWPrivatePS -f $fwPrivateProfileState | Out-File -Append $uninstallLogFile
    $msg.infoFWPublicPS -f $fwPublicProfileState | Out-File -Append $uninstallLogFile
    "" | Tee-Object -FilePath $uninstallLogFile -Append
    $msg.infoRunning -f $scriptName, "" | Tee-Object -FilePath $uninstallLogFile -Append
    "" | Tee-Object -FilePath $uninstallLogFile -Append
}
function StartPsqlServices {
    ##########################
    # If stopped, starts any service with a display name beginning with the string
    # identified by $CmnSvcDispName.
    ##########################
    $msg.doStartPsqlSvcs | Tee-Object -FilePath $uninstallLogFile -Append
    $foundSvc=$false
    get-service | where {$_.DisplayName -Like $CmnSvcDispName+"*"} | Sort-Object -Descending |
        ForEach-Object -Process {
            $foundSvc=$true
            $svcDisplayName=$_.DisplayName
            if ($_.Status -ne "Running") {
                Start-Service -DisplayName $svcDisplayName -ErrorAction Ignore | Out-Null
                if ($?) {
                    $msg.passStartSvcs -f $svcDisplayName | Tee-Object -FilePath $uninstallLogFile -Append
                    # Wait 5 seconds.
                    Start-Sleep -Milliseconds 5000
                } else {
                    $msg.warnStopSvcs -f $svcDisplayName | Tee-Object -FilePath $uninstallLogFile -Append
                }
            } else {
                $msg.skipStartSvcs -f $svcDisplayName | Tee-Object -FilePath $uninstallLogFile -Append
            }
        }
    if (!$foundSvc) {$msg.notFoundSvcs -f $CmnSvcDispName | Tee-Object -FilePath $uninstallLogFile -Append}
    "" | Tee-Object -FilePath $uninstallLogFile -Append
}
function InvokeECollect {
    ##########################
    # Wrapper function for making DTI call pvCollectLocal() to collect the
    # PSQL product information during uninstall.  This call requires
    # the appropriate ELS library is installed so we skip this step if the
    # library is not found.
    #
    # API Invoked
    #   32-bit systems:
    #      clientlm.dll::pvCollectLocal(COLLECT_ON_UNINSTALL)
    #   64-bit systems:
    #      w64clientlm.dll::pvCollectLocal(COLLECT_ON_UNINSTALL)
    ##########################

    # Skip eCollect if the appropriate ELS library is not present or if this is
    # a client or reporting engine install (pvCollectLocal only supported during engine installs).
    if ($Is64Bit) { $elsLib=$PsqlInstDir64 + "bin\w64els.dll"} else {$elsLib=$PsqlInstDir32 + "bin\w3els.dll"}
    if (!(Test-Path $elsLib) -or $isClient  -or $isReporting) {
        $msg.skipECollect | Tee-Object -FilePath $uninstallLogFile -Append
        "" | Tee-Object -FilePath $uninstallLogFile -Append
        return
    }

    $msg.doECollect | Tee-Object -FilePath $uninstallLogFile -Append
    if ($Is64Bit) {$dllName="w64clientlm.dll"} else {$dllName="clientlm.dll"}
    [int]$eventId=110

    # Create the eCollect ScriptBlock. We need to invoke the ScriptBlock as a
    # background task because PowerShell has no way of unloading a DLL after the
    # Add-Type cmdlet has been used, resulting in w64/clientlm.dll and all
    # dependent libraries being locked.
    # Note the two lines in the 'here string' that identify the ELS library
    # signature can't have any leading spaces.
    $apiScriptBlock= {
        param ([string]$dllFileName="clientlm.dll", [int]$eCollectEventId=100)
        $ApiSignature = @"
[DllImport("$dllFileName")]
public static extern int pvCollectLocal(int nEventId);
"@
        Try {
            $pvCollectLocal= Add-Type -MemberDefinition $ApiSignature -Name "Win32pvCollectLocal" -Namespace Win32Functions -PassThru
            $pvCollectLocal::pvCollectLocal($eCollectEventId)
        }
        Catch {[string]$error[0]}
    }

    # Execute the scriptblock with the appropriate library name and event ID.
    Try {
        Start-Job -Name eCollect -ScriptBlock $apiScriptBlock -args $dllName,$eventId | Out-Null
        Wait-Job -Name eCollect | Out-Null
        $result=Receive-Job -Name eCollect
        if ($result.GetType().ToString() -eq "System.Int32") {
            # ScriptBlock returned int.    
            if ($result -eq 0) {
                $msg.passECollect | Tee-Object -FilePath $uninstallLogFile -Append
            } else {
                $msg.warnECollect -f "pvCollectLocal call",$result | Tee-Object -FilePath $uninstallLogFile -Append
            }
        } else {
           # ScriptBlock returned non-int.
           $msg.warnECollect -f "eCollect ScriptBlock",$result | Tee-Object -FilePath $uninstallLogFile -Append
        }
    }
    Catch {$msg.warnECollect -f "eCollect ScriptBlock",$result | Tee-Object -FilePath $uninstallLogFile -Append}
    # If $PsqlDataDir\logs\els.log exists copy it to $env:temp\psql_els_uninstall.log
    if (Test-Path "$PsqlDataDir\logs\els.log") {
        Copy-Item "$PsqlDataDir\logs\els.log" -Destination "$env:temp\psql_els_uninstall.log"
        if ($?) { $msg.infoECollectCopiedLog -f $env:temp | Out-File -Append $uninstallLogFile }
    }
    "" | Tee-Object -FilePath $uninstallLogFile -Append
}
function UnloadPerfMonCounterSets {
    ##########################
    # Wrapper for Unlodctr.exe, called to unload the PSQL counter sets for
    # Windows Performance Monitor.  This is only run during Server Engine,
    # Vx Server, or Reporting Engine uninstalls and only if unlodctr.exe is
    # found.
    #
    # Executes:
    #   "$env:windir\system32\unlodctr.exe" /m:{path}\{XML config file}
    ##########################
    # Skip if WGE or Client install.
    if ( $isWorkgroup -or $isClient) {
        # -- Windows Performance Monitor counter sets are not installed for WGE or Client.
        return
    }
    # Skip if system doesn't support it (no unlodctr.exe).
    $unloader="$env:windir\system32\unlodctr.exe"
    if (!(Test-Path $unloader)) {return}

    # Continue (unlodctr.exe found and SE, Vx, or Rpt engine).
    $msg.doUpdatePerfMonCtr | Tee-Object -FilePath $uninstallLogFile -Append
    ## Set the configuration file.
    if ($isServer -or $isVxServer) {
        if ($Is64Bit) {
            $perfctrcfg=$PsqlInstDir64 + "bin\psqlperfctrSE64.xml"
        } else {
            $perfctrcfg=$PsqlInstDir32 + "bin\psqlperfctrSE32.xml"
        }
    } else {
        if ($isReporting) { $perfctrcfg=$PsqlInstDir64 + "bin\psqlperfctrRE64.xml" }
    }
    if (Test-Path $perfctrcfg) {
        Try {
            $result=& UnlodCtr.exe /m:"$perfctrcfg" | Out-String -Stream
            if ($LASTEXITCODE -eq 0) {
                $msg.passUpdatePerfMonCtr | Tee-Object -FilePath $uninstallLogFile -Append
            } else {
                $msg.warnUpdatePerfMonCtr -f $perfctrcfg, [string]$LASTEXITCODE, [string]$result | Tee-Object -FilePath $uninstallLogFile -Append
            }
        }
        Catch { $msg.warnUpdatePerfMonCtr -f $perfctrcfg, [string]$LASTEXITCODE, [string]$result | Tee-Object -FilePath $uninstallLogFile -Append }
    } else {
        $msg.warnNotFoundPerfMonCtrCfg | Tee-Object -FilePath $uninstallLogFile -Append
    }
    "" | Tee-Object -FilePath $uninstallLogFile -Append
}
function RemovePsqlDatabases {
    ##########################
    # Removes the PSQL system databases created during installation using the same configuration
    # file (specified by $setupdbCfgInstall).
    #
    # PREREQ: The PSQL Service(s) must be started prior to calling setupdb.exe.
    ##########################
    if ( $isReporting -or $isClient) {
        # -- No databases are created during install for client or reporting engine.
        $msg.skipRemoveDbs | Tee-Object -FilePath $uninstallLogFile -Append
        "" | Tee-Object -FilePath $uninstallLogFile -Append
        return
    }

    # -- Continue (Server, Vx, or Workgroup only).
    $msg.doRemoveDbs | Tee-Object -FilePath $uninstallLogFile -Append
    $setupdbLog=$env:TEMP + "\PSQL_uninstall_Setupdb.log"
    $setupdb_args="-r","-log",$setupdbLog,"-dir",$PsqlDataDir,$setupdbCfgInstall
    if ($Is64Bit) {
        if (Test-Path $PsqlInstDir64"bin\setupdbcli.exe") {
            # -- No return code from setupdbcli.exe so just log command line.
            "Executing 'setupdbcli.exe $setupdb_args'." | Tee-Object -FilePath $uninstallLogFile -Append
            Try {
                Start-Process -FilePath $PsqlInstDir64"bin\setupdbcli.exe" -ArgumentList $setupdb_args -WorkingDirectory $PsqlInstDir64"bin" -NoNewWindow -Wait
                if ($?) { $msg.passSetupdb -f $PsqlInstDir64, $setupdbLog | Tee-Object -FilePath $uninstallLogFile -Append }
            }
            Catch { $msg.failSetupdb -f $PsqlInstDir64 | Tee-Object -FilePath $uninstallLogFile -Append }

        } else {
            $msg.noSetupDbRemove -f $PsqlInstDir64 | Tee-Object -FilePath $uninstallLogFile -Append
        }
    } else {
        if (Test-Path $PsqlInstDir32"bin\setupdbcli.exe") {
            # -- No return code from setupdbcli.exe so just log command line.
            "Executing 'setupdbcli.exe $setupdb_args'." | Tee-Object -FilePath $uninstallLogFile -Append
            Try {
                Start-Process -FilePath $PsqlInstDir32"bin\setupdbcli.exe" -ArgumentList $setupdb_args -WorkingDirectory $PsqlInstDir32"bin" -NoNewWindow -Wait
                if ($?) { $msg.passSetupdb -f $PsqlInstDir32, $setupdbLog | Tee-Object -FilePath $uninstallLogFile -Append }
            }
            Catch { $msg.failSetupdb -f $PsqlInstDir32 | Tee-Object -FilePath $uninstallLogFile -Append }
        } else {
            $msg.noSetupDbRemove -f $PsqlInstDir32 | Tee-Object -FilePath $uninstallLogFile -Append
        }
    }
    "" | Tee-Object -FilePath $uninstallLogFile -Append
}
function StopPsqlServices {
    ##########################
    # Stop any running service with a display name beginning with the string
    # identified by $CmnSvcDispName.
    ##########################
    $msg.doStopSvcs -f $CmnSvcDispName | Tee-Object -FilePath $uninstallLogFile -Append
    $foundSvc=$false
    get-service -ErrorAction Ignore | where {$_.DisplayName -Like $CmnSvcDispName+"*"} |
        ForEach-Object -Process {
            if ($_.Status -eq "Running") {
                $foundSvc=$true
                $svcDisplayName=$_.DisplayName
                Stop-Service -DisplayName $svcDisplayName -Force -ErrorAction Ignore | Out-Null
                if ($?) {
                    $msg.passStopSvcs -f $svcDisplayName | Tee-Object -FilePath $uninstallLogFile -Append
                } else {
                    $msg.warnStopSvcs -f $svcDisplayName | Tee-Object -FilePath $uninstallLogFile -Append
                }
                # Wait 5 seconds.
                Start-Sleep -Milliseconds 5000
            }
        }
    if (!$foundSvc) {$msg.notFoundRunningSvcs -f $CmnSvcDispName | Tee-Object -FilePath $uninstallLogFile -Append}
    "" | Tee-Object -FilePath $uninstallLogFile -Append
}
function RemovePsqlFirewallRules {
    ##########################
    # Removes any existing PSQL firewall rule if NetSecurity module exists.
    ##########################
    if ( $isClient) {
        # -- No firewall rules are created for client installs.
        return
    }
    $moduleNetSecurity=$false
    Get-Command -CommandType Function | ForEach-Object -Process {if ($_.Source -eq "NetSecurity") {$moduleNetSecurity=$true} }
    if ($moduleNetSecurity) {
        Import-Module -Name NetSecurity  # -- Firewall rule cmdlets.
    } else {
        # -- The NetSecurity Module is not installed, warn user and exit function.
        $msg.warnNoNetSecRemove | Tee-Object -FilePath $uninstallLogFile -Append
        "" | Tee-Object -FilePath $uninstallLogFile -Append
        return
    }
    $msg.doRemoveFWRules | Tee-Object -FilePath $uninstallLogFile -Append
    Remove-NetFirewallRule -DisplayName "$CmnSvcDispName*" #| Out-Null
    if ($?) {
        $msg.passRemoveFWRule -f $CmnSvcDispName | Tee-Object -FilePath $uninstallLogFile -Append
    } else {
        $msg.warnRemoveFWRule -f $CmnSvcDispName | Tee-Object -FilePath $uninstallLogFile -Append
    }
    "" | Tee-Object -FilePath $uninstallLogFile -Append
}
function RemovePsqlServices {
    ##########################
    # Use the Service Controller utility to delete any stopped service with a display name
    # that begins witht he string specified by $CmnSvcDispName.
    ##########################
    if ( $isClient) {
        # -- No services are created for client installs.
        return
    }
    $msg.doRemoveSvc | Tee-Object -FilePath $uninstallLogFile -Append
    get-service -ErrorAction Ignore | where {$_.DisplayName -Like $CmnSvcDispName+"*"} |
        ForEach-Object -Process {
            if ($_.Status -eq "Stopped") {
                $svcName=$_.Name
                $svcDisplayName=$_.DisplayName
                "Executing 'sc.exe delete $svcName'" | Tee-Object -FilePath $uninstallLogFile -Append
                $scReturn=& sc.exe delete $svcName
                if ($scReturn.Contains("DeleteService SUCCESS")) {
                    $msg.passRemoveSvc -f $svcDisplayName | Tee-Object -FilePath $uninstallLogFile -Append

                } else {
                    $msg.warnUnableToStopSvc -f $svcDisplayName, [string]$scReturn | Tee-Object -FilePath $uninstallLogFile -Append
                }
            } else {
               $msg.warnSvcNotStopped -f $svcDisplayName | Tee-Object -FilePath $uninstallLogFile -Append
            }
        }
    "" | Tee-Object -FilePath $uninstallLogFile -Append
}
function UnregisterComInterfaces {
    ##########################
    # Checks to see which PSQL COM interface library files are installed and then unregisters
    # each file using the regsvr32.exe utility.
    ##########################
    $msg.doUnRegisterCom  | Tee-Object -FilePath $uninstallLogFile -Append
    $nFound=0;
    $nUnregistered=0;
    if ($Is64Bit) {$filePath=$PsqlInstDir64} else {$filePath=$PsqlInstDir32}
    foreach ($fileToRegister in $ComFileList) {
        $fullFilePath=$filePath+"bin\"+$fileToRegister
        if (Test-Path $fullFilePath) {
            $nFound++
            & regsvr32.exe -s -u "$fullFilePath"| Out-String -Stream
            switch ($LASTEXITCODE) {
                0       {$msg.passUnRegFile -f $fullFilePath | Out-File -Append $uninstallLogFile; $nUnregistered++;break}
                default {$msg.warnUnRegFile -f $fullFilePath | Tee-Object -FilePath $uninstallLogFile -Append; break}
            }
        }
    }
    "Successfully un-registered $nUnregistered out of $nFound files. Refer to $uninstallLogFile for more information." | Tee-Object -FilePath $uninstallLogFile -Append
    "" | Tee-Object -FilePath $uninstallLogFile -Append
}
function RemoveStrFromPath {
    ##########################
    # Removes the specified string from the PATH environment variable.
    #   $args[0]: Path string to be removed from %PATH%
    ##########################
    if ($args.Count -ne 1) {
        $msg.warnRemoveFromPathArgs | Tee-Object -FilePath $uninstallLogFile -Append
         return
    } else {
        $msg.doRemoveFromPath -f $args[0] | Tee-Object -FilePath $uninstallLogFile -Append
        $match1="(;" + $args[0] + ")" -replace '(\\)','$1\'  # reg exp pattern 1 to match=(;path)
        $match2="(" + $args[0] + ";)" -replace '(\\)','$1\'  # reg exp pattern 2 to match=(path;)
        # -- Remove from string from the Local Machine (System) path.
        $regPathParent="HKLM:System\CurrentControlSet\Control\Session Manager\Environment"
        $currentMachinePath=Get-ItemPropertyValue -Path $regPathParent -Name "Path"
        $newMachinePath=$currentMachinePath -replace "$match1|$match2","" # replace either matching pattern with empty string
        Set-ItemProperty -Path $regPathParent -Name "Path" -Value $newMachinePath
        if ($?) {
            $msg.passRemoveFromPath -f $args[0], "machine" | Tee-Object -FilePath $uninstallLogFile -Append
        } else {
            $msg.warnRemoveFromPath -f $args[0], "machine" | Tee-Object -FilePath $uninstallLogFile -Append
        }
        # -- Remove from string from the current PowerShell session path.
        $currentSessionPath=[System.Environment]::GetEnvironmentVariable("Path")
        $newSessionPath=$currentSessionPath -replace "$match1|$match2","" # replace either matching pattern with empty string
        [Environment]::SetEnvironmentVariable("Path", $newSessionPath)
        if ($?) {
            $msg.passRemoveFromPath -f $args[0], "current session" | Tee-Object -FilePath $uninstallLogFile -Append
        } else {
            $msg.warnRemoveFromPath -f $args[0], "current session" | Tee-Object -FilePath $uninstallLogFile -Append
        }
        "" | Tee-Object -FilePath $uninstallLogFile -Append
    }
}
function RemovePsqlRegistry {
    ##########################
    # Removes the root PSQL registry key and all sub-keys.
    ##########################
    $msg.doRemovePsqlReg -f $registryPath  | Tee-Object -FilePath $uninstallLogFile -Append
    ## Root key.
    if (Test-Path $registryPath) {
        Try {
            Remove-Item $registryPath -Recurse
            if ($?) {$msg.passRemovePsqlReg -f $registryPath | Tee-Object -FilePath $uninstallLogFile -Append}
        }
        Catch {$msg.warnRemovePsqlReg -f $registryPath | Tee-Object -FilePath $uninstallLogFile -Append}
    } else {$msg.notFoundPsqlReg -f $registryPath | Tee-Object -FilePath $uninstallLogFile -Append}
    "" | Tee-Object -FilePath $uninstallLogFile -Append
}
function RemoveDirAndParentIfEmpty {
    ##########################
    # Deletes the specified folder if it exists and is empty and then checks to see
    # if the parent folder is now empty and removes it if it is empty.
    #   $args[0]: Folder path to be removed.
    ##########################
    if ($args.Count -ne 1) {
        $msg.warnRemoveDirArgs | Tee-Object -FilePath $uninstallLogFile -Append
         return
    } else {
        $dirObject=Get-Item $args[0]
        if (Test-Path $dirObject.FullName) {
            # -- Remove folder if it is empty.
            if ( ($dirObject.GetDirectories("*").Count -eq 0) -and ($dirObject.GetFiles("*").Count -eq 0) ) {
                Try {
                    Remove-Item $dirObject.FullName -Force -ErrorVariable e -ErrorAction SilentlyContinue
                    if ($?) {
                        $msg.passRemoveDir -f $dirObject.FullName | Tee-Object -FilePath $uninstallLogFile -Append
                    } else {
                        $msg.warnRemoveDir -f $dirObject.FullName, [string]$e | Tee-Object -FilePath $uninstallLogFile -Append
                    }
                }
                Catch {$msg.warnRemoveDir -f $dirObject.FullName, [string]$error[0] | Tee-Object -FilePath $uninstallLogFile -Append}
             }
             # -- Remove parent folder if it is now empty.
             if ( ($dirObject.Parent.GetDirectories("*").Count -eq 0) -and ($dirObject.Parent.GetDirectories("*").Count -eq 0) ) {
                 Try {
                     Remove-Item $dirObject.Parent.FullName -Force -ErrorVariable e -ErrorAction SilentlyContinue
                     if ($?) {
                         $msg.passRemoveParent -f $dirObject.Parent.FullName | Tee-Object -FilePath $uninstallLogFile -Append
                     } else {
                         $msg.warnRemoveDir -f $dirObject.FullName, [string]$e | Tee-Object -FilePath $uninstallLogFile -Append
                     }
                 }
                 Catch {$msg.warnRemoveDir -f $dirObject.FullName, [string]$error[0] | Tee-Object -FilePath $uninstallLogFile -Append}
             } else {
                 $msg.warnRemoveParent -f $dirObject.Parent.FullName | Tee-Object -FilePath $uninstallLogFile -Append
             }
        }
    }
}
function RemovePsqlFiles {
    ##########################
    # Removes all installed PSQL files except for any existing DBNAMES.cfg.
    ##########################
    $msg.doRemovePsqlFiles  | Tee-Object -FilePath $uninstallLogFile -Append
    # -- Program Files
    if ( ($PsqlInstDir64) -and (Test-Path $PsqlInstDir64) ) {
        $objDir=Get-Item $PsqlInstDir64 -ErrorAction Ignore
        Try {
            Remove-Item $PsqlInstDir64 -Recurse -Force -ErrorVariable e -ErrorAction SilentlyContinue
            if ($?) {
                $msg.passRemovePsqlFiles -f $PsqlInstDir64 | Tee-Object -FilePath $uninstallLogFile -Append
                RemoveDirAndParentIfEmpty $objDir.Parent.FullName
            } else {
                $msg.warnRemovePsqlFiles -f $PsqlInstDir64, [string]$e | Tee-Object -FilePath $uninstallLogFile -Append
            }
        }
        Catch {$msg.warnRemovePsqlFiles -f $PsqlInstDir64, [string]$error[0] | Tee-Object -FilePath $uninstallLogFile -Append}
    }
    # -- Only remove $PsqlInstDir32 if it is not the same as $PsqlInstDir64 (they are the same on Nano Server systems).
    if ( ($PsqlInstDir64 -ne $PsqlInstDir32) -and ($PsqlInstDir32) -and (Test-Path $PsqlInstDir32) ) {
        $objDir=Get-Item $PsqlInstDir32 -ErrorAction Ignore
        Try {
            Remove-Item $PsqlInstDir32 -Recurse -Force -ErrorVariable e -ErrorAction SilentlyContinue
            if ($?) {
                $msg.passRemovePsqlFiles -f $PsqlInstDir32 | Tee-Object -FilePath $uninstallLogFile -Append
                RemoveDirAndParentIfEmpty $ObjDir.Parent.FullName
            } else {
                $msg.warnRemovePsqlFiles -f $PsqlInstDir32, [string]$e | Tee-Object -FilePath $uninstallLogFile -Append
            }
        }
        Catch {$msg.warnRemovePsqlFiles -f $PsqlInstDir32, [string]$error[0] | Tee-Object -FilePath $uninstallLogFile -Append}
    }
    # -- For Data Files we need to keep the DBNAMES.CFG file and the defaultdb, Demodata, system, and tempdb folders.
    $safeList="DBNAMES.CFG","defaultdb","Demodata","system","tempdb"
    if (Test-Path $PsqlDataDir"\DBNAMES.CFG") {
        # -- If DBNAMES.CFG exists, only delete items not in $safeList.     
        Get-ChildItem $PsqlDataDir | ForEach-Object -Process {
            $itemToDelete=$_.Name
            if ($safeList -icontains $itemToDelete) {
                # ---Do nothing, item must be kept.
                $msg.keepDbnames -f $PsqlDataDir, $itemToDelete | Tee-Object -FilePath $uninstallLogFile -Append  
            } else {
                Try {
                    Remove-Item $PsqlDataDir$itemToDelete -Recurse -Force -ErrorVariable e -ErrorAction SilentlyContinue
                    if ($?) {
                        $msg.passRemovePsqlItem -f $PsqlDataDir, $itemToDelete | Tee-Object -FilePath $uninstallLogFile -Append
                    } else {
                        $msg.warnRemovePsqlItem -f $PsqlDataDir, $itemToDelete | Tee-Object -FilePath $uninstallLogFile -Append
                    }
                }
                Catch {$msg.warnRemovePsqlItem -f $PsqlDataDir, $itemToDelete | Tee-Object -FilePath $uninstallLogFile -Append}
            }
        }
    } else {
        # -- If DBNAMES.CFG does NOT exist, remove $PsqlDataDir altogether.
        if ( ($PsqlDataDir) -and (Test-Path $PsqlDataDir) ) {
            $objDir=Get-Item $PsqlDataDir -ErrorAction Ignore
            Try {
                Remove-Item $PsqlDataDir -Recurse -Force -ErrorVariable e -ErrorAction SilentlyContinue
                if ($?) {
                    $msg.passRemovePsqlFiles -f $PsqlDataDir  | Tee-Object -FilePath $uninstallLogFile -Append
                    RemoveDirAndParentIfEmpty $objDir.Parent.FullName
                } else {
                    $msg.warnRemovePsqlFiles -f $PsqlDataDir, [string]$e | Tee-Object -FilePath $uninstallLogFile -Append
                }
            }
            Catch {$msg.warnRemovePsqlFiles -f $PsqlDataDir, [string]$error[0] | Tee-Object -FilePath $uninstallLogFile -Append}
        }
    }
    "" | Tee-Object -FilePath $uninstallLogFile -Append
}
function EndLog {
    ##########################
    # Close $uninstallLogFile with date/time script completed.
    $msg.infoEnd -f [DateTime]::Parse((Get-Date).ToString()) | Out-File -Append $uninstallLogFile
    ##########################
    "#############################" | Tee-Object -FilePath $uninstallLogFile -Append
    "" | Out-File -Append $uninstallLogFile
    $msg.completeRemoveMsg -f $uninstallLogFile
    "" | Tee-Object -FilePath $uninstallLogFile -Append
}

Try {. Main}
Catch {throw $msg.errUnknown; exit -1}
# SIG # Begin signature block
# MIIcMwYJKoZIhvcNAQcCoIIcJDCCHCACAQExCzAJBgUrDgMCGgUAMGkGCisGAQQB
# gjcCAQSgWzBZMDQGCisGAQQBgjcCAR4wJgIDAQAABBAfzDtgWUsITrck0sYpfvNR
# AgEAAgEAAgEAAgEAAgEAMCEwCQYFKw4DAhoFAAQU+4IwDIRN9bDR0KWjrz7lvPOQ
# +T2gghcYMIID7jCCA1egAwIBAgIQfpPr+3zGTlnqS5p31Ab8OzANBgkqhkiG9w0B
# AQUFADCBizELMAkGA1UEBhMCWkExFTATBgNVBAgTDFdlc3Rlcm4gQ2FwZTEUMBIG
# A1UEBxMLRHVyYmFudmlsbGUxDzANBgNVBAoTBlRoYXd0ZTEdMBsGA1UECxMUVGhh
# d3RlIENlcnRpZmljYXRpb24xHzAdBgNVBAMTFlRoYXd0ZSBUaW1lc3RhbXBpbmcg
# Q0EwHhcNMTIxMjIxMDAwMDAwWhcNMjAxMjMwMjM1OTU5WjBeMQswCQYDVQQGEwJV
# UzEdMBsGA1UEChMUU3ltYW50ZWMgQ29ycG9yYXRpb24xMDAuBgNVBAMTJ1N5bWFu
# dGVjIFRpbWUgU3RhbXBpbmcgU2VydmljZXMgQ0EgLSBHMjCCASIwDQYJKoZIhvcN
# AQEBBQADggEPADCCAQoCggEBALGss0lUS5ccEgrYJXmRIlcqb9y4JsRDc2vCvy5Q
# WvsUwnaOQwElQ7Sh4kX06Ld7w3TMIte0lAAC903tv7S3RCRrzV9FO9FEzkMScxeC
# i2m0K8uZHqxyGyZNcR+xMd37UWECU6aq9UksBXhFpS+JzueZ5/6M4lc/PcaS3Er4
# ezPkeQr78HWIQZz/xQNRmarXbJ+TaYdlKYOFwmAUxMjJOxTawIHwHw103pIiq8r3
# +3R8J+b3Sht/p8OeLa6K6qbmqicWfWH3mHERvOJQoUvlXfrlDqcsn6plINPYlujI
# fKVOSET/GeJEB5IL12iEgF1qeGRFzWBGflTBE3zFefHJwXECAwEAAaOB+jCB9zAd
# BgNVHQ4EFgQUX5r1blzMzHSa1N197z/b7EyALt0wMgYIKwYBBQUHAQEEJjAkMCIG
# CCsGAQUFBzABhhZodHRwOi8vb2NzcC50aGF3dGUuY29tMBIGA1UdEwEB/wQIMAYB
# Af8CAQAwPwYDVR0fBDgwNjA0oDKgMIYuaHR0cDovL2NybC50aGF3dGUuY29tL1Ro
# YXd0ZVRpbWVzdGFtcGluZ0NBLmNybDATBgNVHSUEDDAKBggrBgEFBQcDCDAOBgNV
# HQ8BAf8EBAMCAQYwKAYDVR0RBCEwH6QdMBsxGTAXBgNVBAMTEFRpbWVTdGFtcC0y
# MDQ4LTEwDQYJKoZIhvcNAQEFBQADgYEAAwmbj3nvf1kwqu9otfrjCR27T4IGXTdf
# plKfFo3qHJIJRG71betYfDDo+WmNI3MLEm9Hqa45EfgqsZuwGsOO61mWAK3ODE2y
# 0DGmCFwqevzieh1XTKhlGOl5QGIllm7HxzdqgyEIjkHq3dlXPx13SYcqFgZepjhq
# IhKjURmDfrYwggR9MIIDZaADAgECAgMb5xUwDQYJKoZIhvcNAQELBQAwYzELMAkG
# A1UEBhMCVVMxITAfBgNVBAoTGFRoZSBHbyBEYWRkeSBHcm91cCwgSW5jLjExMC8G
# A1UECxMoR28gRGFkZHkgQ2xhc3MgMiBDZXJ0aWZpY2F0aW9uIEF1dGhvcml0eTAe
# Fw0xNDAxMDEwNzAwMDBaFw0zMTA1MzAwNzAwMDBaMIGDMQswCQYDVQQGEwJVUzEQ
# MA4GA1UECBMHQXJpem9uYTETMBEGA1UEBxMKU2NvdHRzZGFsZTEaMBgGA1UEChMR
# R29EYWRkeS5jb20sIEluYy4xMTAvBgNVBAMTKEdvIERhZGR5IFJvb3QgQ2VydGlm
# aWNhdGUgQXV0aG9yaXR5IC0gRzIwggEiMA0GCSqGSIb3DQEBAQUAA4IBDwAwggEK
# AoIBAQC/cWII8fpZNPcbyRij94BJWOkigxOmxSBDATuE8eaFSZ8n6vaEG06gtNtw
# mMcyAbEFPgdO7vT6Ty9ZMCLnqxlWa+KAB/zzFnWAOVF75fk1tnROqY2CE+S2P6kD
# g/qivooVan/eC8O2GRQFyurDqASUO0Z8Mg3zAGYiyI1pbTaMERi307IcYLQ4+gKM
# ztPdRgfeCj7rXXzIfPuwK1OkkmJpUSUFYRpEgYwsqUOWI9+sOoGaDinFHKnpXR62
# np4wCjnO8YiA+0tdzDLshWJDJTQCVicBkbQ7cCo/brHonIgBfZ/U+dtTbWCdvyzn
# WKu4X0b8zsQbAzwJ60kxXGlGs+BHAgMBAAGjggEXMIIBEzAPBgNVHRMBAf8EBTAD
# AQH/MA4GA1UdDwEB/wQEAwIBBjAdBgNVHQ4EFgQUOpqFBxBnKLbv9r0FQW4gwZTa
# D94wHwYDVR0jBBgwFoAU0sSw0pHUTBFxs2HLPaH+3ahq1OMwNAYIKwYBBQUHAQEE
# KDAmMCQGCCsGAQUFBzABhhhodHRwOi8vb2NzcC5nb2RhZGR5LmNvbS8wMgYDVR0f
# BCswKTAnoCWgI4YhaHR0cDovL2NybC5nb2RhZGR5LmNvbS9nZHJvb3QuY3JsMEYG
# A1UdIAQ/MD0wOwYEVR0gADAzMDEGCCsGAQUFBwIBFiVodHRwczovL2NlcnRzLmdv
# ZGFkZHkuY29tL3JlcG9zaXRvcnkvMA0GCSqGSIb3DQEBCwUAA4IBAQBZC1O9koYR
# pyR77Vsxzx0fbHDFuG6+Trv2vpdQ4TB/uihcYpTC434z9/tCdoXblRyMIlh1CQyI
# ZWc5ChYJxaA4l6TFI5M/tBimAQZEkeOnaSe0WiV/Orcyzd2E/yo4KTOk3Weyhf6h
# iCAcUInI3Cr2QgM3TOaI39WvJPKxw9/MtezgmV63SVQgPJQYDMccUhhJpG3hs1gL
# ydjs2a4cMo4ocA3i/qYXnoQPvVdws1rpH6CGU7vvfP9pC+BIw7eTC8gKVMSsXRRn
# N2zKpS8xCDeqbm+MvJviV10kga+Xl5yErWysN0xm82GRESDkvjCfeqQpCbDhNF9k
# dxhAUd+MMKavMIIEozCCA4ugAwIBAgIQDs/0OMj+vzVuBNhqmBsaUDANBgkqhkiG
# 9w0BAQUFADBeMQswCQYDVQQGEwJVUzEdMBsGA1UEChMUU3ltYW50ZWMgQ29ycG9y
# YXRpb24xMDAuBgNVBAMTJ1N5bWFudGVjIFRpbWUgU3RhbXBpbmcgU2VydmljZXMg
# Q0EgLSBHMjAeFw0xMjEwMTgwMDAwMDBaFw0yMDEyMjkyMzU5NTlaMGIxCzAJBgNV
# BAYTAlVTMR0wGwYDVQQKExRTeW1hbnRlYyBDb3Jwb3JhdGlvbjE0MDIGA1UEAxMr
# U3ltYW50ZWMgVGltZSBTdGFtcGluZyBTZXJ2aWNlcyBTaWduZXIgLSBHNDCCASIw
# DQYJKoZIhvcNAQEBBQADggEPADCCAQoCggEBAKJjCzlEuLsjp0RJuw7/ofBhClOT
# sJjbrSwPSsVu/4Y8U1UPFc4EPyv9qZaW2b5heQtbyUyGduXgQ0sile7CK0PBn9ho
# tI5AT+6FOLkRxSPyZFjwFTJvTlehroikAtcqHs1L4d1j1ReJMluwXplaqJ0oUA4X
# 7pbbYTtFUR3PElYLkkf8q672Zj1HrHBy55LnX80QucSDZJQZvSWA4ejSIqXQugJ6
# oXeTW2XD7hd0vEGGKtwITIySjJEtnndEH2jWqHR32w5bMotWizO92WPISZ06xcXq
# MwvS8aMb9Iu+2bNXizveBKd6IrIkri7HcMW+ToMmCPsLvalPmQjhEChyqs0CAwEA
# AaOCAVcwggFTMAwGA1UdEwEB/wQCMAAwFgYDVR0lAQH/BAwwCgYIKwYBBQUHAwgw
# DgYDVR0PAQH/BAQDAgeAMHMGCCsGAQUFBwEBBGcwZTAqBggrBgEFBQcwAYYeaHR0
# cDovL3RzLW9jc3Aud3Muc3ltYW50ZWMuY29tMDcGCCsGAQUFBzAChitodHRwOi8v
# dHMtYWlhLndzLnN5bWFudGVjLmNvbS90c3MtY2EtZzIuY2VyMDwGA1UdHwQ1MDMw
# MaAvoC2GK2h0dHA6Ly90cy1jcmwud3Muc3ltYW50ZWMuY29tL3Rzcy1jYS1nMi5j
# cmwwKAYDVR0RBCEwH6QdMBsxGTAXBgNVBAMTEFRpbWVTdGFtcC0yMDQ4LTIwHQYD
# VR0OBBYEFEbGaaMOShQe1UzaUmMXP142vA3mMB8GA1UdIwQYMBaAFF+a9W5czMx0
# mtTdfe8/2+xMgC7dMA0GCSqGSIb3DQEBBQUAA4IBAQB4O7SRKgBM8I9iMDd4o4Qn
# B28Yst4l3KDUlAOqhk4ln5pAAxzdzuN5yyFoBtq2MrRtv/QsJmMz5ElkbQ3mw2cO
# 9wWkNWx8iRbG6bLfsundIMZxD82VdNy2XN69Nx9DeOZ4tc0oBCCjqvFLxIgpkQ6A
# 0RH83Vx2bk9eDkVGQW4NsOo4mrE62glxEPwcebSAe6xp9P2ctgwWK/F/Wwk9m1vi
# FsoTgW0ALjgNqCmPLOGy9FqpAa8VnCwvSRvbIrvD/niUUcOGsYKIXfA9tFGheTMr
# Lnu53CAJE3Hrahlbz+ilMFcsiUk/uc9/yb8+ImhjU5q9aXSsxR08f5Lgw7wc2AR1
# MIIE0DCCA7igAwIBAgIBBzANBgkqhkiG9w0BAQsFADCBgzELMAkGA1UEBhMCVVMx
# EDAOBgNVBAgTB0FyaXpvbmExEzARBgNVBAcTClNjb3R0c2RhbGUxGjAYBgNVBAoT
# EUdvRGFkZHkuY29tLCBJbmMuMTEwLwYDVQQDEyhHbyBEYWRkeSBSb290IENlcnRp
# ZmljYXRlIEF1dGhvcml0eSAtIEcyMB4XDTExMDUwMzA3MDAwMFoXDTMxMDUwMzA3
# MDAwMFowgbQxCzAJBgNVBAYTAlVTMRAwDgYDVQQIEwdBcml6b25hMRMwEQYDVQQH
# EwpTY290dHNkYWxlMRowGAYDVQQKExFHb0RhZGR5LmNvbSwgSW5jLjEtMCsGA1UE
# CxMkaHR0cDovL2NlcnRzLmdvZGFkZHkuY29tL3JlcG9zaXRvcnkvMTMwMQYDVQQD
# EypHbyBEYWRkeSBTZWN1cmUgQ2VydGlmaWNhdGUgQXV0aG9yaXR5IC0gRzIwggEi
# MA0GCSqGSIb3DQEBAQUAA4IBDwAwggEKAoIBAQC54MsQ1K92vdSTYuswZLiBCGzD
# BNliF44v/z5lz4/OYuY8UhzaFkVLVat4a2ODYpDOD2lsmcgaFItMzEUz6ojcnqOv
# K/6AYZ15V8TPLvQ/MDxdR/yaFrzDN5ZBUY4RS1T4KL7QjL7wMDge87Am+GZHY23e
# cSZHjzhHU9FGHbTj3ADqRay9vHHZqm8A29vNMDp5T19MR/gd71vCxJ1gO7GyQ5HY
# pDNO6rPWJ0+tJYqlxvTV0KaudAVkV4i1RFXULSo6Pvi4vekyCgKUZMQWOlDxSq7n
# eTOvDCAHf+jfBDnCaQJsY1L6d8EbyHSHyLmTGFBUNUtpTrw700kuH9zB0lL7AgMB
# AAGjggEaMIIBFjAPBgNVHRMBAf8EBTADAQH/MA4GA1UdDwEB/wQEAwIBBjAdBgNV
# HQ4EFgQUQMK9J47MNIMwojPX+2yz8LQsgM4wHwYDVR0jBBgwFoAUOpqFBxBnKLbv
# 9r0FQW4gwZTaD94wNAYIKwYBBQUHAQEEKDAmMCQGCCsGAQUFBzABhhhodHRwOi8v
# b2NzcC5nb2RhZGR5LmNvbS8wNQYDVR0fBC4wLDAqoCigJoYkaHR0cDovL2NybC5n
# b2RhZGR5LmNvbS9nZHJvb3QtZzIuY3JsMEYGA1UdIAQ/MD0wOwYEVR0gADAzMDEG
# CCsGAQUFBwIBFiVodHRwczovL2NlcnRzLmdvZGFkZHkuY29tL3JlcG9zaXRvcnkv
# MA0GCSqGSIb3DQEBCwUAA4IBAQAIfmyTEMg4uJapkEv/oV9PBO9sPpyIBslQj6Zz
# 91cxG7685C/b+LrTW+C05+Z5Yg4MotdqY3MxtfWoSKQ7CC2iXZDXtHwlTxFWMMS2
# RJ17LJ3lXubvDGGqv+QqG+6EnriDfcFDzkSnE3ANkR/0yBOtg2DZ2HKocyQetawi
# DsoXiWJYRBuriSUBAA/NxBti21G00w9RKpv0vHP8ds42pM3Z2Czqrpv1KrKQ0U11
# GIo/ikGQI31bS/6kA1ibRrLDYGCD+H1QQc7CoZDDu+8CL9IVVO5EFdkKrqeKM+2x
# LXY2JtwE65/3YR8V3Idv7kaWKK2hJn0KCacuBKONvPi8BDABMIIFJjCCBA6gAwIB
# AgIICsxCB6AJxjswDQYJKoZIhvcNAQELBQAwgbQxCzAJBgNVBAYTAlVTMRAwDgYD
# VQQIEwdBcml6b25hMRMwEQYDVQQHEwpTY290dHNkYWxlMRowGAYDVQQKExFHb0Rh
# ZGR5LmNvbSwgSW5jLjEtMCsGA1UECxMkaHR0cDovL2NlcnRzLmdvZGFkZHkuY29t
# L3JlcG9zaXRvcnkvMTMwMQYDVQQDEypHbyBEYWRkeSBTZWN1cmUgQ2VydGlmaWNh
# dGUgQXV0aG9yaXR5IC0gRzIwHhcNMTcwNDA1MjAwMjAwWhcNMjAwNDEwMTk1MTM5
# WjBoMQswCQYDVQQGEwJVUzEOMAwGA1UECBMFVGV4YXMxDzANBgNVBAcTBkF1c3Rp
# bjEbMBkGA1UEChMSQWN0aWFuIENvcnBvcmF0aW9uMRswGQYDVQQDExJBY3RpYW4g
# Q29ycG9yYXRpb24wggEiMA0GCSqGSIb3DQEBAQUAA4IBDwAwggEKAoIBAQDDl5bZ
# gsKcnWzrY8rYrsSmI4l9tC5Ipk1Dp/LHYZQ/1uwgzbiUQavCOKhJ6q+XMGHIOF6P
# N/PkfIfrf1XygR3bJy+f0u4HICioNg0GmfBgEUHVpZTkeRxCT+Dd1FeDDljp9J41
# AVkXh8SnzKvSGKjaUbp3DkH2Y5P24pHR9zhpx2hnqSF0PHOT0fGFcF4SAeYBTIT0
# nSmVPxjs2+NV1M81xtCLLLCF9ceV784DyY7MhoHI9Q7fHpxUJZi0nuHRzyUurofC
# QmJhOLd5vm7re/WUAHfhG3P7dIkm1mI9SHAWid2A/4vGWHb/bR5mwJTDxvPmgfzz
# EVmR8PHPNTsRXZYpAgMBAAGjggGFMIIBgTAMBgNVHRMBAf8EAjAAMBMGA1UdJQQM
# MAoGCCsGAQUFBwMDMA4GA1UdDwEB/wQEAwIHgDA1BgNVHR8ELjAsMCqgKKAmhiRo
# dHRwOi8vY3JsLmdvZGFkZHkuY29tL2dkaWcyczUtMi5jcmwwXQYDVR0gBFYwVDBI
# BgtghkgBhv1tAQcXAjA5MDcGCCsGAQUFBwIBFitodHRwOi8vY2VydGlmaWNhdGVz
# LmdvZGFkZHkuY29tL3JlcG9zaXRvcnkvMAgGBmeBDAEEATB2BggrBgEFBQcBAQRq
# MGgwJAYIKwYBBQUHMAGGGGh0dHA6Ly9vY3NwLmdvZGFkZHkuY29tLzBABggrBgEF
# BQcwAoY0aHR0cDovL2NlcnRpZmljYXRlcy5nb2RhZGR5LmNvbS9yZXBvc2l0b3J5
# L2dkaWcyLmNydDAfBgNVHSMEGDAWgBRAwr0njsw0gzCiM9f7bLPwtCyAzjAdBgNV
# HQ4EFgQUd28uU+LsLPArCXvwS49GX5IaKcUwDQYJKoZIhvcNAQELBQADggEBAKMz
# CNStFW8r/jaaSDXwbWnpmRjeGx3FZbmwH+r2gv2UQ5Se0ckdRm86CfF8AjGnFd/d
# pOSsLWUSWx5+gdqorB9+xY6KrB0br08CfRYvvw+ZYiRrnvLEJyD3hc13tHYpmDOS
# lamg3seFzk/rKe5g/ZXwDQXOiAMEkH5xBHS3BBILQarbiJ14CLM9eLDSeN8L/Gdy
# SvMX4nWy9UusiXP3ZXvuVPUlAOBvDUZaM1BPir0ZWY/m2FOW8bHxXzTG8VtdgY0y
# JIkaii7O3A1kkB9ofB47LQN/CuwcOOKQx0hNL3Nlq7zsFizWk6MO3ZmPNjjimkI6
# mT0NuoY6yd9VefcO95MxggSFMIIEgQIBATCBwTCBtDELMAkGA1UEBhMCVVMxEDAO
# BgNVBAgTB0FyaXpvbmExEzARBgNVBAcTClNjb3R0c2RhbGUxGjAYBgNVBAoTEUdv
# RGFkZHkuY29tLCBJbmMuMS0wKwYDVQQLEyRodHRwOi8vY2VydHMuZ29kYWRkeS5j
# b20vcmVwb3NpdG9yeS8xMzAxBgNVBAMTKkdvIERhZGR5IFNlY3VyZSBDZXJ0aWZp
# Y2F0ZSBBdXRob3JpdHkgLSBHMgIICsxCB6AJxjswCQYFKw4DAhoFAKCBijAZBgkq
# hkiG9w0BCQMxDAYKKwYBBAGCNwIBBDAcBgorBgEEAYI3AgELMQ4wDAYKKwYBBAGC
# NwIBFTAjBgkqhkiG9w0BCQQxFgQU8oArIq2IMUTcVJ7w/x1D/H+ICq8wKgYKKwYB
# BAGCNwIBDDEcMBqhGIAWaHR0cDpcXHd3dy5hY3RpYW4uY29tIDANBgkqhkiG9w0B
# AQEFAASCAQBofZffDVdDiTCD6cJLiD+sK17DPdX75gerbaFdF/YLHtqMPt2CZAZq
# p98YWHLbairUhBl01THIyJYF9I1IzYXoC44jpfawtOSyO42887zwXbXPXw/fbDY3
# fssqmjgAzLtbvNsnAXbwFHWzXSGM49qObDigerB5OcW1uFLZe1E0D3mMWYA9aeZ6
# GrPj1sQ5EUcqMNphWYBsy4ppSxGtDfJ9KHV0V2NXMqHbLV9PIzR3txDZOZ8IYpB+
# ltsvGxuag3wiFTdQ0xz8P1H2u5TkHLkisC5GqAdLsUnE5i1QRnwf9K4We0p7yrSV
# ecpIzNgfYhYU/Ct4yE+Iaz701jw/IM2roYICCzCCAgcGCSqGSIb3DQEJBjGCAfgw
# ggH0AgEBMHIwXjELMAkGA1UEBhMCVVMxHTAbBgNVBAoTFFN5bWFudGVjIENvcnBv
# cmF0aW9uMTAwLgYDVQQDEydTeW1hbnRlYyBUaW1lIFN0YW1waW5nIFNlcnZpY2Vz
# IENBIC0gRzICEA7P9DjI/r81bgTYapgbGlAwCQYFKw4DAhoFAKBdMBgGCSqGSIb3
# DQEJAzELBgkqhkiG9w0BBwEwHAYJKoZIhvcNAQkFMQ8XDTE4MTAyMzE1MzA1Nlow
# IwYJKoZIhvcNAQkEMRYEFI60eR5rN5bmBK6/W99whgWDAWYNMA0GCSqGSIb3DQEB
# AQUABIIBAGpwM4OGh8IJBESYnX5LKOefxyOapYPFY80/zo+zSy+VrHaMp0mlWFwf
# fcVvISklFZ5vvI1A5AX7xS8zJ5pkZgwe3YP/+DAbePNbyEUg52XNEMvuOFAxyQxm
# 1LHBVyB/+RrrpZnCHBPHqc2zsHAE5c6G6GfuOzaiyM5AInXwhw8o9uZZ9+cL6PRv
# RgzVqayUBO7Pqm8WK+sKh7jGA9/FGACKHv7sleKiDDMeuesL+aRtngtTJXPfi1B2
# 2u0pdSu+dxNl0qSSqbpBimtA+53TUCJDwos92fVTv6t8uDG5e/fVOcSiQQaUlexU
# 8Xnz9bkZTL6YF7GZw1H01zGAODu9YRk=
# SIG # End signature block
