##############################################################################
##
## Install-PSQL
##
## by Actian Corporation (http://www.actian.com/psql)
##
##############################################################################

    <#
    .NOTES
        Copyright (c) Actian Corporation.  All rights reserved.

        Use of this sample source code is subject to the terms of the Actian
        PSQL license agreement under which you licensed this sample source code.
        If you did not accept the terms of the license agreement, you are not
        authorized to use this sample source code. For the terms of the license,
        please see the license agreement between you and Actian Corporation.
        THE SAMPLE SOURCE CODE IS PROVIDED "AS IS", WITH NO WARRANTIES.

    .SYNOPSIS
        Installs PSQL on a local system running Windows Nano Server or Windows
        IoT Core.

    .DESCRIPTION
        Installs PSQL on a local system running either 64-bit Windows Nano Server
        or 32-bit Windows Internet of Things (IoT).  The user must specify one of
        the parameters specified below.  The script can be invoked multiple times
        to update existing files and/or to add new files.

    .PARAMETER server
        Installs PSQL Server Engine.  On 64-bit Windows Nano Server, the 64-bit
        PSQL Server Engine is installed.  On 32-bit Windows IoT, the 32-bit
        PSQL Server Engine is installed.

    .PARAMETER client
        Only available on Windows Nano Server.  Installs PSQL Client files for
        accessing a PSQL database on a remote system.

    .PARAMETER reporting
        Only available on Windows Nano Server.  Installs the 64-bit PSQL
        Reporting Engine.

    .EXAMPLE
        PS C:\> .\Install-PSQL server
            Install the PSQL Server Engine.

    .EXAMPLE
        PS C:\> .\Install-PSQL client
            Install the PSQL Client (Windows Nano Server only).

    .EXAMPLE
        PS C:\> .\Install-PSQL reporting
            Installs the PSQL Reporting Engine (Windows Nano Server only).

    .INPUTS
        None. You cannot pipe objects to this cmdlet.

    .OUTPUTS
        0 : The script completed with no errors (may contain warnings).
        -1: The script returned one or more errors.

        Creates log file in users $env:temp folder.
        Creates object $PsqlInfo with the following properties:
                $PsqlInfo.InstallData    : Install location of data files.
                $PsqlInfo.InstallDir64   : Install location on 64-bit systems.
                $PsqlInfo.InstallDir     : Install location on 32-bit systems,
                                           same as InstallDir64 on 64-bit systems.
                $PsqlInfo.IsReport       : workgroup (0) or reporting (1) only
                $PsqlInfo.IsVx           : server (0) or vx (1) only.
                $PsqlInfo.ProductVersion : The full PSQL major.minor.build version.
                $PsqlInfo.VersionLevel   : The PSQL major.minor version.

    #>

#Requires -Version 5.1        # -- Require PowerShell 5.1 or higher (Nano Server,IoT).
#Requires -PSEdition Core     # -- Require PowerShell Core
#Requires -RunAsAdministrator # -- Require PowerShell running as Administrator.

param (
    [Parameter(
        Mandatory=$true,
        Position = 0)]
    [ValidateSet('server','client','reporting')]
    [String]
    $InstallProductType
)

Set-StrictMode -Version Latest

Import-LocalizedData msg -FileName "PSQL_PowerShell_Script_strings.psd1" -ErrorAction Ignore

#############################
## Variables Required prior to calling SetScriptVariables().
#############################
if ([System.IntPtr]::Size -eq 4) {$script:Is64Bit=$false} else {$script:Is64Bit=$true}
$script:scriptPath = & {Split-Path $myInvocation.ScriptName}
$script:registryPath = "HKLM:\SOFTWARE\Pervasive Software"
$script:installing=$true
$script:updating=$false
switch ($InstallProductType) {
    'client' { $script:InstallInfoProdType="Pervasive.SQL Client"; break}
    'server' { $script:InstallInfoProdType="Pervasive.SQL NT Server"; break}
    'vx' { $script:InstallInfoProdType="Pervasive.SQL NT Server"; break}
    'workgroup' { $script:InstallInfoProdType="Pervasive.SQL Workgroup"; break}
    'reporting' { $script:InstallInfoProdType="Pervasive.SQL Workgroup"; break}
}
$script:culture=Get-Culture
if ( $PSCulture.StartsWith("ja") ) {$script:IsJajp=$true} else {$script:IsJajp=$false}

#############################
## Abort Checks Not Using #Requires
#############################
# -- Abort Reporting Engine or Client install if OS is 32-bit.
if (!$Is64Bit -and ($InstallProductType.Equals("reporting") ) ) { throw $msg.errNoRPT32; exit -1 }
# -- Abort if other product type detected.
$abortInstall=$false
if (test-path $registryPath"\Products\") {
    # -- Found existing "Products" key, verify that any existing sub-keys that begin "Pervasive.SQL"
    #    end with a string matching $InstallInfoProdType (see SetScriptVariables).
    #    Assumption #1: For Server/Vx, we allow install of Server if Vx already installed (and vice
    #                   versa) so no need to check InstallInfo\IsVx value.
    #    Assumption #2: Workgroup and Reporting are already mutually exclusive based on the OS
    #                   bit-ness checks above so no need to check InstallInfo\IsReport value.
    $alreadyInstalled=""
    Get-ChildItem $registryPath"\Products\" | ForEach-Object -Process {
        if ($_.Name -Like "*Pervasive.SQL*" ) {
            if ($_.Name.EndsWith("$installInfoProdType",1)) {
                $updating=$true
            } else {
                $abortInstall=$true
                $alreadyInstalled=$_.Name
            }
        }
    }
}
if ($abortInstall) { throw $msg.errNoOtherPsql -f $InstallProductType, $alreadyInstalled; exit -1}
# -- Abort if install_$InstallProductType not found.
if (Test-Path -Path $scriptPath"\install_"$InstallProductType".cfg") {
} else {
    throw $msg.errNoProductType -f "$scriptPath\install_$InstallProductType.cfg"; exit -1
}
# -- Abort if ProgramFiles ZIP in package doesn't match this platform.
if (Test-Path -Path "$scriptPath\ProgramFilesX64.zip") {$script:expectedPlatform="x86_64"}
if (Test-Path -Path "$scriptPath\ProgramFilesX86.zip") {$script:expectedPlatform="x86"}
if (Test-Path -Path "$scriptPath\ProgramFilesArm32.zip") {$script:expectedPlatform="32-bit ARM"}
if ($env:PROCESSOR_ARCHITECTURE -eq "ARM") {
    # ARM processor, only 32-bit supported.
    if ($Is64Bit) { throw $msg.errNoArm64; exit -1 } else {
        if ($expectedPlatform -ne "32-bit ARM") { throw $msg.errWrongPlatform -f $expectedPlatform,"32-bit ARM"; exit -1}
    }
} else {
    if ($Is64Bit) {
        if ($expectedPlatform -ne "x86_64") { throw $msg.errWrongPlatform -f $expectedPlatform,"x86_64"; exit -1}
    } else {
        if ($expectedPlatform -ne "x86") { throw $msg.errWrongPlatform -f $expectedPlatform,"x86"; exit -1}
    }
}
#############################
## FUNCTIONS
#############################
function Main {
    ##########################
    # Defines the processing order of this script.
    ##########################
    SetScriptVariables
    BeginLog
    ReadFromConfig
    if ($updating) {UnloadPerfMonCounterSets}
    StopPsqlServices
    UnpackFiles
    RemoveFilesByProductType
    UpdatePath
    InitPsqlRegistry
    RegisterComInterfaces
    CreatePsqlServices
    AddPsqlFirewallRules
    StartPsqlServices
    ActivatePsqlTempLicense
    CreatePsqlDatabases
    UpdateRoyaltyRecord
    InvokeECollect
    LoadPerfMonCounterSets
    EndLog
}
function SetScriptVariables {
    ##########################
    # Sets script-level variables based on specified product type to be installed. Except
    # for $PsqlInstallCfg, of the variable is set to a simple string then it can be
    # overridden by specifying the name=value pair in the install configuration file
    # specified by $PsqlInstallCfg.
    #
    # Note: Must be called before logging is started.
    ##########################
    $script:PsqlInstallCfg="install_"+ $InstallProductType + ".cfg"
    $script:defaultAclArgs="Everyone","FullControl","ContainerInherit","InheritOnly","Allow"
    $script:scriptName = & {$myInvocation.ScriptName}
    #$script:scriptPath = & {Split-Path $myInvocation.ScriptName}  ## - set before Abort Checks section.
    $script:computerName = ${env:computerName}
    [string]$fwpDomainState=netsh advfirewall show domainprofile State
    if ( $fwpDomainState.Contains("ON") ) {$script:fwDomainProfileState="ON"} else {$script:fwDomainProfileState="OFF"}
    [string]$fwpPrivateState=netsh advfirewall show privateprofile State
    if ( $fwpPrivateState.Contains("ON") ) {$script:fwPrivateProfileState="ON"} else {$script:fwPrivateProfileState="OFF"}
    [string]$fwpPublicState=netsh advfirewall show publicprofile State
    if ( $fwpPublicState.Contains("ON") ) {$script:fwPublicProfileState="ON"} else {$script:fwPublicProfileState="OFF"}
    #$script:registryPath = "HKLM:\SOFTWARE\Pervasive Software"  ## - Set before Abort Checks section.
    $script:CmnSvcDispName = "Actian PSQL"
    # -- Default ELS settings.
    $script:psqlElsUrl=""
    $script:psqlElsServerID="0"
    $script:fpCommonApplicationData=${env:ALLUSERSPROFILE}
    if ($Is64Bit) {
        $script:fpProgramFiles64=${env:ProgramFiles}
        $script:fpProgramFiles32=${env:ProgramFiles(x86)}
    } else {
        $script:fpProgramFiles32=${env:ProgramFiles}
        $script:fpProgramFiles64=$fpProgramFiles32
    }
    # -- Duplicating MSI behavior, folder paths end with a backslash.
    $script:PsqlDataDir="$fpCommonApplicationData\Actian\PSQL\"
    $script:PsqlInstDir64="$fpProgramFiles64\Actian\PSQL\"
    $script:PsqlInstDir32="$fpProgramFiles32\Actian\PSQL\"
    $script:setupdbCfgInstall=$PsqlDataDir + "w32setupdb.cfg"
    # -- Default values (overwritten based on $InstallProductType)
    $script:isClient=$False
    $script:isServer=$False
    $script:isVxServer=$False
    $script:isWorkgroup=$False
    $script:isReporting=$False
    $script:installLogFile = $env:TEMP + "\PSQL_PowerShell_Install.log"
    # -- Variables based on command line arugments.
    switch ($InstallProductType) {
        'client' {
            $script:isClient=$True
            $script:InstallInfoProdType="Pervasive.SQL Client"
            $script:installLogFile = $env:TEMP + "\PSQL_Client_PowerShell_Install.log"
            $script:filesToRemoveFromBin = "w3dbsmgr.exe","uwmkde.dll","w64dbsmgr.exe","w64repmkde.dll","ntbtrv.exe","ntdbsmgr.exe","ntmkde.dll","w3sqlmgr.exe","ntbtrv64.exe","ntdbsmgr64.exe","w64ntmkde.dll","w64sqlmgr.exe","psqlperfctr*.xml"
            $script:filesToRemoveFromPsql = ""
            $script:filesToRemoveFromData = "collate.cfg","defaultdb","Demodata","icudt54l.dat","samples","system","Tempdb","u54-msft_enus_0.txt","w32setupdb.cfg"
            break}
        'server' {
            $script:isServer=$True
            $script:InstallInfoProdType="Pervasive.SQL NT Server"
            $script:installLogFile = $env:TEMP + "\PSQL_ServerEngine_PowerShell_Install.log"
            $script:filesToRemoveFromBin = "w3dbsmgr.exe","uwmkde.dll","w64dbsmgr.exe","w64repmkde.dll","psqlperfctrRE64.xml"
            $script:filesToRemoveFromPsql = ""
            $script:filesToRemoveFromData = ""
            break}
        'vx' {
            $script:isVxServer=$True
            $script:InstallInfoProdType="Pervasive.SQL NT Server"
            $script:installLogFile = $env:TEMP + "\PSQL_VxServer_PowerShell_Install.log"
            $script:filesToRemoveFromBin = "w3dbsmgr.exe","uwmkde.dll","w64dbsmgr.exe","w64repmkde.dll","psqlperfctrRE64.xml"
            $script:filesToRemoveFromPsql = ""
            $script:filesToRemoveFromData = ""
            break}
        'workgroup' {
            $script:isWorkgroup=$True
            $script:InstallInfoProdType="Pervasive.SQL Workgroup"
            $script:installLogFile = $env:TEMP + "\PSQL_WorkgroupEngine_PowerShell_Install.log"
            $script:filesToRemoveFromBin = "w64dbsmgr.exe","w64repmkde.dll","ntbtrv.exe","ntdbsmgr.exe","ntmkde.dll","w3sqlmgr.exe","ntbtrv64.exe","ntdbsmgr64.exe","w64ntmkde.dll","w64sqlmgr.exe","psqlperfctr*.xml"
            $script:filesToRemoveFromPsql = ""
            $script:filesToRemoveFromData = ""
            break}
        'reporting' {
            $script:isReporting=$True
            $script:InstallInfoProdType="Pervasive.SQL Workgroup"
            $script:installLogFile = $env:TEMP + "\PSQL_ReportingEngine_PowerShell_Install.log"
            $script:filesToRemoveFromBin = "dbdefrag.exe","w3dbsmgr.exe","uwmkde.dll","ntbtrv.exe","ntdbsmgr.exe","ntmkde.dll","w3sqlmgr.exe","ntbtrv64.exe","ntdbsmgr64.exe","w64ntmkde.dll","w64sqlmgr.exe","psqlperfctrSE*.xml"
            $script:filesToRemoveFromPsql = ""
            $script:filesToRemoveFromData = "defaultdb","Demodata","samples","system","w32setupdb.cfg"
            break}
    }
    # -- COM self-registering file lists.
    $script:ComFileList="acbtr732.ocx","acctr732.ocx","clientrb.dll","cobolschemaexecmsgrb.dll","dbcsipxy.dll","dto.dll","dto2.dll","legacylm.dll","licmgrrb.dll","mkc3.dll","mkderb.dll","pceurop.dll","pctlgrb.dll","poledb.dll","poledber.dll","poledbsy.dll","pscp932.dll","pscp936.dll","pscp949.dll","pscp950.dll","pscrb_en.dll","pscrb_ja.dll","psdom.dll","pseucjp.dll","pssax.dll","pstestct.dll","pstestsp.dll","pstsprb.dll","pvconnts_en.dll","pvconnts_ja.dll","pvinfodg_en.dll","pvinfodg_ja.dll","pvmsgrb.dll","upiapirb.dll","w3cobolschemaexec100.dll","w3csi100.dll","w3csm100.dll","w3csp100.dll","w3dcm100.dll","w3exp010.dll","w3mpm100.dll","w64clientrb.dll","w64cobolschemaexec100.dll","w64cobolschemaexecmsgrb.dll","w64csi.dll","w64csm.dll","w64csp.dll","w64dbcsipxy.dll","w64dcm.dll","w64dto2.dll","w64exp010.dll","w64legacylm.dll","w64licmgrrb.dll","w64mkc3.dll","w64mkderb.dll","w64mpm.dll","w64pceurop.dll","w64pctlgrb.dll","w64pscp932.dll","w64pscp936.dll","w64pscp949.dll","w64pscp950.dll","w64psdom.dll","w64pseucjp.dll","w64pssax.dll","w64pvmsgrb.dll","w64upiapirb.dll"
}
function BeginLog {
    ##########################
    # Create $installLogFile and write basic info to the file.
    ##########################
    "#############################" | Out-File -Append $installLogFile
    $msg.infoStart -f [DateTime]::Parse((Get-Date).ToString()) | Out-File -Append $installLogFile
    $msg.infoScriptName -f $scriptName | Out-File -Append $installLogFile
    $msg.infoScriptPath -f $ScriptPath | Out-File -Append $installLogFile
    $msg.infoUpdating -f $updating | Out-File -Append $installLogFile
    $msg.infoCompName -f $computerName | Out-File -Append $installLogFile
    $msg.infoCulture -f $culture.Name | Out-File -Append $installLogFile
    $msg.infoProcNum -f [System.Environment]::ProcessorCount | Out-File -Append $installLogFile
    $msg.infoWinArch -f $env:PROCESSOR_ARCHITECTURE | Out-File -Append $installLogFile
    $msg.infoWinPlat -f [System.Environment]::OSVersion.Platform | Out-File -Append $installLogFile
    $msg.infoWinVer -f [System.Environment]::OSVersion.Version | Out-File -Append $installLogFile
    $msg.infoIs64Bit -f $Is64Bit | Out-File -Append $installLogFile
    $msg.infoCLRVer -f [System.Environment]::Version | Out-File -Append $installLogFile
    $msg.infoUserName -f $env:USERNAME | Out-File -Append $installLogFile
    $msg.infoUserDomain -f $env:USERDOMAIN | Out-File -Append $installLogFile
    $msg.infoPSVer -f $psVersionTable.PSVersion | Out-File -Append $installLogFile
    $msg.infoPSCLRVer -f $psVersionTable.CLRVersion | Out-File -Append $installLogFile
    $msg.infoFWDomainPS -f $fwDomainProfileState | Out-File -Append $installLogFile
    $msg.infoFWPrivatePS -f $fwPrivateProfileState | Out-File -Append $installLogFile
    $msg.infoFWPublicPS -f $fwPublicProfileState | Out-File -Append $installLogFile
    "" | Tee-Object -FilePath $installLogFile -Append
    $msg.infoRunning -f $scriptName, $InstallProductType | Tee-Object -FilePath $installLogFile -Append
    "" | Tee-Object -FilePath $installLogFile -Append
}
function StopPsqlServices {
    ##########################
    # Stop any running service with a display name beginning with the string
    # identified by $CmnSvcDispName.
    ##########################
    $msg.doStopSvcs -f $CmnSvcDispName | Tee-Object -FilePath $installLogFile -Append
    $foundSvc=$false
    get-service -ErrorAction Ignore | where {$_.DisplayName -Like $CmnSvcDispName+"*"} |
        ForEach-Object -Process {
            if ($_.Status -eq "Running") {
                $foundSvc=$true
                $svcDisplayName=$_.DisplayName
                Stop-Service -DisplayName $svcDisplayName -Force -ErrorAction Ignore | Out-Null
                if ($?) {
                    $msg.passStopSvcs -f $svcDisplayName | Tee-Object -FilePath $installLogFile -Append
                } else {
                    $msg.warnStopSvcs -f $svcDisplayName | Tee-Object -FilePath $installLogFile -Append
                }
                # Wait 5 seconds.
                Start-Sleep -Milliseconds 5000
            }
        }
    if (!$foundSvc) {$msg.notFoundRunningSvcs -f $CmnSvcDispName | Tee-Object -FilePath $installLogFile -Append}
    "" | Tee-Object -FilePath $installLogFile -Append
}
function UnpackFiles {
    ##########################
    # Unpacks the ZIP files associated with this script to the appropriate default Windows
    # location and then calls UpdateACLs to set the ACL for the folder to allow full
    # access for everyone.
    ##########################
    $msg.doUnpack | Tee-Object -FilePath $installLogFile -Append
    ## - Always uncompress DataFiles.zip.
    if (Test-Path -Path "$scriptPath\DataFiles.zip") {
        Expand-Archive -PATH "$scriptPath\DataFiles.zip" -DestinationPath $fpCommonApplicationData -Force
        if ($?) {
            $msg.passUnpack -f "$scriptPath\DataFiles.zip", $fpCommonApplicationData | Tee-Object -FilePath $installLogFile -Append
            UpdateACLs $PsqlDataDir $defaultAclArgs
        } else {
            $msg.warnUnpack -f "$scriptPath\DataFiles.zip", $fpCommonApplicationData | Tee-Object -FilePath $installLogFile -Append
        }
    }
    ## - Program Files depends on bitness and architecture.
    if ($env:PROCESSOR_ARCHITECTURE -eq "ARM") {
        ## -- Windows on ARM is currently 32-bit only.
        if ($Is64Bit) { throw $msg.errNoArm64; exit -1 }
        $extractTo=$fpProgramFiles32
        $PgmFilesZip="$scriptPath\ProgramFilesArm32.zip"
        $PgmFilesJaZip="$scriptPath\ProgramFilesArm32_jajp.zip"
    } else {
        if ($Is64Bit) {
            ## -- 64-bit Windows.
            $extractTo=$fpProgramFiles64
            $PgmFilesZip="$scriptPath\ProgramFilesX64.zip"
            $PgmFilesJaZip="$scriptPath\ProgramFilesX64_jajp.zip"
        } else {
            ## -- 32-bit Windows.
            $extractTo=$fpProgramFiles32
            $PgmFilesZip="$scriptPath\ProgramFilesX86.zip"
            $PgmFilesJaZip="$scriptPath\ProgramFilesX86_jajp.zip"
        }
    }
    if (Test-Path -Path "$PgmFilesZip") {
        Expand-Archive -Path "$PgmFilesZip" -DestinationPath $extractTo -Force
        if ($?) {
            $msg.passUnpack -f "$PgmFilesZip", $extractTo | Tee-Object -FilePath $installLogFile -Append
            UpdateACLs "$extractTo\Actian" $defaultAclArgs
        } else {
            $msg.warnUnpack -f "$PgmFilesZip", $extractTo | Tee-Object -FilePath $installLogFile -Append
        }
    }
    if ($IsJajp) {
        if (Test-Path -Path "$PgmFilesJaZip") {
            Expand-Archive -Path "$PgmFilesJaZip" -DestinationPath $extractTo -Force
            if ($?) {
                $msg.passUnpack -f "$PgmFilesJaZip", $extractTo | Tee-Object -FilePath $installLogFile -Append
            } else {
                $msg.warnUnpack -f "$PgmFilesJaZip", $extractTo | Tee-Object -FilePath $installLogFile -Append
            }
        }
    }
    "" | Tee-Object -FilePath $installLogFile -Append
}
function RemoveFilesByProductType {
    ##########################
    # For each product type, the list of files identified by the file lists below
    # are removed from the folders identified below:
    #    $filesToRemoveFromBin files are removed from the Bin folder.
    #    $filesToRemoveFromPsql files are removed from the PSQL folder (bin parent).
    #    $filesToRemoveFromData files are removed from the Data folder.
    ##########################
    $msg.doDelFiles -f $InstallProductType | Out-File -Append $installLogFile
    if ($Is64Bit) {$binDir=$PsqlInstDir64+"bin"} else {$binDir=$PsqlInstDir32+"bin"}
    if ($filesToRemoveFromBin -ne "") {
        $filesToRemoveFromBin | Foreach-Object -Process {
            if (Test-Path "$binDir\$_") {
                Try {
                    Remove-Item -Path "$binDir\$_" -Force
                    if ($?) { $msg.passDelFiles -f $InstallProductType, "$binDir\$_" | Out-File -Append $installLogFile }
                }
                Catch { $msg.warnDelFiles -f $InstallProductType, "$binDir\$_" | Out-File -Append $installLogFile }
            }
        }
    }
    if ($Is64Bit) {$psqlDir=$PsqlInstDir64} else {$psqlDir=$PsqlInstDir32}
    if ($filesToRemoveFromPsql -ne "") {
        $filesToRemoveFromPsql | Foreach-Object -Process {
            if (Test-Path "$psqlDir\$_") {
                Try {
                    Remove-Item -Path "$psqlDir\$_" -Force
                    if ($?) { $msg.passDelFiles -f $InstallProductType, "$psqlDir\$_" | Out-File -Append $installLogFile }
                }
                Catch { $msg.warnDelFiles -f $InstallProductType, "$psqlDir\$_" | Out-File -Append $installLogFile }
            }
        }
    }
    if ($filesToRemoveFromData -ne "") {
        $filesToRemoveFromData | Foreach-Object -Process {
            if (Test-Path $PsqlDataDir$_) {
                Try {
                    Remove-Item -Path $PsqlDataDir$_ -Recurse -Force
                    if ($?) { $msg.passDelFiles -f $InstallProductType, "$PsqlDataDir$_" | Out-File -Append $installLogFile }
                }
                Catch { $msg.warnDelFiles -f $InstallProductType, "$PsqlDataDir$_" | Out-File -Append $installLogFile }
            }
        }
    }
    "" | Out-File -Append $installLogFile
}
function UpdateACLs {
    ##########################
    # Updates the ACL for the specified file, folder, or registry key using the Get-Acl and
    # Set-Acl cmdlets and the .Net constructors for FileSystemAccessRule or RegistryAccessRule
    # to update the current ACLs to the specified ACLs according to the input parameters:
    # Assumes arguments are specified in the following order:
    #   $args[0]: Path to file, folder or registry key.
    #   $args[1]: ACL settings to be passed to the FileSystemAccessRule or RegistryAccessRule
    #             .net constructors.
    ##########################
    $ruleType="FileSystemAccessRule"
    if ($args.Count -ne 2) {
        $msg.warnUpdateACLs1 | Tee-Object -FilePath $installLogFile -Append
        return
    } else {
        $pathToUpdate=[string]$args[0]
        $aclArgs=$args[1]
    }
    $msg.doUpdateACLs -f [string]$aclArgs, $pathToUpdate | Tee-Object -FilePath $installLogFile -Append
    if (Test-Path $args[0]) {
        if ($pathToUpdate.IndexOf(":") -ge 4) {$ruleType="RegistryAccessRule"}
        try {
            $acl=Get-Acl $pathToUpdate
            $accessRule = New-Object System.Security.AccessControl.$ruleType $aclArgs
            $acl.SetAccessRule($accessRule)
            $acl | Set-Acl $pathToUpdate
            if ($?) { $msg.passUpdateACLs -f [string]$aclArgs, $pathToUpdate | Tee-Object -FilePath $installLogFile -Append }
        } catch {
            $msg.warnUpdateACLs3 -f $pathToUpdate, $aclArgs | Tee-Object -FilePath $installLogFile -Append
        }
    } else {
        $msg.warnUpdateACLs2 -f $args[0], $args[1] | Tee-Object -FilePath $installLogFile -Append
    }
    "" | Tee-Object -FilePath $installLogFile -Append
}
function ReadFromConfig {
    ##########################
    # Reads the content of $PsqlInstallCfg (same dir as script) and defines a global variable
    # for each specified name=value pair. If the variable is already defined it is
    # updated.
    ##########################
    $msg.doReadCfgFile -f $PsqlInstallCfg | Tee-Object -FilePath $installLogFile -Append
    if (Test-Path "$scriptPath\$PsqlInstallCfg") {
        # Set script-level properties.
        $props = Get-Content "$scriptPath\$PsqlInstallCfg"
        For ($i=0; $i -lt $props.Count; $i++) {
            if ( $props[$i].StartsWith("#") ) {
                # Skip comment line.
            } else {
                if ( $props[$i].Contains("=") ) {
                    $items=$props[$i] -split "="
                    # Check if variable already set
                    Get-Variable $items[0] -ErrorAction Ignore | Out-Null
                    if ($?) {
                        $msg.addVarFromCfg -f $items[0], $items[1] | Out-File -Append $installLogFile
                        Set-Variable $items[0] -Value $items[1] -Scope Script #-errorAction Ignore
                    } else {
                        $msg.updateVarFromCfg -f $items[0], $items[1] | Out-File -Append $installLogFile
                        New-Variable $items[0] -Value $items[1] -Scope Script #-errorAction Ignore
                    }
                }
            }
        }
    } else {
        throw $msg.errNoCfgFile -f "$scriptPath\$PsqlInstallCfg" | Tee-Object -FilePath $installLogFile -Append
        exit -1
    }
    "" | Tee-Object -FilePath $installLogFile -Append
}
function UnloadPerfMonCounterSets {
    ##########################
    # Wrapper for Unlodctr.exe, called to unload the PSQL counter sets for
    # Windows Performance Monitor.  This is only run during Server Engine,
    # Vx Server, or Reporting Engine uninstalls and only if unlodctr.exe is
    # found.
    #
    # Executes:
    #   "$env:windir\system32\unlodctr.exe" /m:{path}\{XML config file}
    ##########################
    # Skip if WGE or Client install.
    if ( $isWorkgroup -or $isClient) {
        # -- Windows Performance Monitor counter sets are not installed for WGE or Client.
        return
    }
    # Skip if system doesn't support it (no unlodctr.exe).
    $unloader="$env:windir\system32\unlodctr.exe"
    if (!(Test-Path $unloader)) {return}

    # Continue (unlodctr.exe found and SE, Vx, or Rpt engine).
    $msg.doUpdatePerfMonCtr | Tee-Object -FilePath $installLogFile -Append
    ## Set the configuration file.
    if ($isServer -or $isVxServer) {
        if ($Is64Bit) {
            $perfctrcfg=$PsqlInstDir64 + "bin\psqlperfctrSE64.xml"
        } else {
            $perfctrcfg=$PsqlInstDir32 + "bin\psqlperfctrSE32.xml"
        }
    } else {
        if ($isReporting) { $perfctrcfg=$PsqlInstDir64 + "bin\psqlperfctrRE64.xml" }
    }
    if (Test-Path $perfctrcfg) {
        Try {
            $result=& UnlodCtr.exe /m:"$perfctrcfg" | Out-String -Stream
            if ($LASTEXITCODE -eq 0) {
                $msg.passUpdatePerfMonCtr | Tee-Object -FilePath $installLogFile -Append
            } else {
                $msg.warnUpdatePerfMonCtr -f $perfctrcfg, [string]$LASTEXITCODE, [string]$result | Tee-Object -FilePath $installLogFile -Append
            }
        }
        Catch { $msg.warnUpdatePerfMonCtr -f $perfctrcfg, [string]$LASTEXITCODE, [string]$result | Tee-Object -FilePath $installLogFile -Append }
    } else {
        $msg.warnNotFoundPerfMonCtrCfg | Tee-Object -FilePath $installLogFile -Append
    }
    "" | Tee-Object -FilePath $installLogFile -Append
}
function UpdatePath {
    ##########################
    # Checks the PATH environment variable to determine if it already contains the
    # PSQL bin folder path.  If not, the PSQL bin folder path is added to the
    # beginning of the environment variable.
    #
    # Note: To keep consistent with the folder paths created by the MSI install, all
    #       folder paths added to the %PATH% below end with a backslash '\'.  If this
    #       is ever changed, the associated function in the uninstall script will
    #       also need to be modified.
    ##########################
    $msg.doUpdatePath | Tee-Object -FilePath $installLogFile -Append
    if ($Is64Bit) {$addToPath=$PsqlInstDir64 + "bin"} else {$addToPath=$PsqlInstDir32 + "bin"}
    # -- Update local machine PATH.
    $regPathParent="HKLM:System\CurrentControlSet\Control\Session Manager\Environment"
    $currentMachinePath=Get-ItemPropertyValue -Path $regPathParent -Name "Path"
    if ($currentMachinePath -like "*$addToPath*") {
        $msg.skipUpdatePath -f "Machine", $addToPath | Tee-Object -FilePath $installLogFile -Append
    } else {
        #Prepend bin location to PATH.
        Set-ItemProperty -Path $regPathParent -Name "Path" -Value "$addToPath\;$currentMachinePath"
        if ($?) {
            $msg.passUpdatePath -f $addToPath, "machine" | Tee-Object -FilePath $installLogFile -Append
        } else {
            $msg.warnUpdatePath -f $addToPath, "machine" | Tee-Object -FilePath $installLogFile -Append
        }
    }
    # -- Update current PowerShell session PATH.
    $currentSessionPath=[System.Environment]::GetEnvironmentVariable("Path")
    if ($currentSessionPath -like "*$addToPath*") {
        $msg.skipUpdatePath -f "Current session", $addToPath | Tee-Object -FilePath $installLogFile -Append
    } else {
        #Prepend bin location to PATH.
        [Environment]::SetEnvironmentVariable("Path", $addToPath +"\;" + $currentSessionPath)
        if ($?) {
            $msg.passUpdatePath -f $addToPath, "current session" | Tee-Object -FilePath $installLogFile -Append
        } else {
            $msg.warnUpdatePath -f $addToPath, "current session" | Tee-Object -FilePath $installLogFile -Append
        }
    }
    "" | Tee-Object -FilePath $installLogFile -Append
}
function InitPsqlRegistry {
    ##########################
    # Create InstallInfo registry key and requisite registry values based on product type
    # specified on command line.  Note that all registry keys and values not explicitly
    # set in this procedure are created with their default values PSQL is started for the
    # first time.
    #
    # Variables set:
    #   [object] $global:PsqlInfo = see .OUTPUT in script header for a list of
    ##########################
    $msg.doPsqlReg  | Tee-Object -FilePath $installLogFile -Append
    ## Root key.
    if (Test-Path $registryPath) {
        $msg.skipCreateRootKey -f $registryPath | Tee-Object -FilePath $installLogFile -Append
    } else {
        $msg.addRootKey -f $registryPath | Tee-Object -FilePath $installLogFile -Append
        New-Item -Path $registryPath -Force | Out-Null
        UpdateACLs $registryPath $defaultAclArgs
    }
    if (Test-Path $registryPath"\Products\"$InstallInfoProdType"\InstallInfo") {
        $msg.skipInstallInfoKey -f $registryPath | Tee-Object -FilePath $installLogFile -Append
    } else {
        $msg.addInstallInfoKey -f $registryPath | Tee-Object -FilePath $installLogFile -Append
        New-Item -Path $registryPath"\Products\"$InstallInfoProdType"\InstallInfo" -Force | Out-Null
    }
    $msg.addInstallInfoValues -f $installLogFile | Tee-Object -FilePath $installLogFile -Append
    New-ItemProperty -Path $registryPath"\Products\"$InstallInfoProdType"\InstallInfo" -Name "InstallData" -Value $PsqlDataDir -PropertyType String -Force | Out-Null
    New-ItemProperty -Path $registryPath"\Products\"$InstallInfoProdType"\InstallInfo" -Name "InstallDir" -Value $PsqlInstDir64 -PropertyType String -Force | Out-Null
    New-ItemProperty -Path $registryPath"\Products\"$InstallInfoProdType"\InstallInfo" -Name "InstallDir64" -Value $PsqlInstDir64 -PropertyType String -Force | Out-Null
    if ($isServer) {New-ItemProperty -Path $registryPath"\Products\"$InstallInfoProdType"\InstallInfo" -Name "IsVx" -Value "0" -PropertyType DWORD -Force | Out-Null}
    if ($isVxServer) {New-ItemProperty -Path $registryPath"\Products\"$InstallInfoProdType"\InstallInfo" -Name "IsVx" -Value "1" -PropertyType DWORD -Force | Out-Null}
    if ($isWorkgroup -or $isReporting) {New-ItemProperty -Path $registryPath"\Products\"$InstallInfoProdType"\InstallInfo" -Name "IsBtrieve" -Value "0" -PropertyType DWORD -Force | Out-Null}
    if ($isWorkgroup) {New-ItemProperty -Path $registryPath"\Products\"$InstallInfoProdType"\InstallInfo" -Name "IsReport" -Value "0" -PropertyType DWORD -Force | Out-Null}
    if ($isReporting) {New-ItemProperty -Path $registryPath"\Products\"$InstallInfoProdType"\InstallInfo" -Name "IsReport" -Value "1" -PropertyType DWORD -Force | Out-Null}
    if ($Is64Bit -and ($isServer -or $isVxServer)) {New-ItemProperty -Path $registryPath"\Products\"$InstallInfoProdType"\InstallInfo" -Name "Server64" -Value "YES" -PropertyType String -Force | Out-Null}
    # -- Use default values if properties not read from configuration file.
    New-ItemProperty -Path $registryPath"\Products\"$InstallInfoProdType"\InstallInfo" -Name "ProductVersion" -Value $ProductVer -PropertyType String -Force | Out-Null
    New-ItemProperty -Path $registryPath"\Products\"$InstallInfoProdType"\InstallInfo" -Name "VersionLevel" -Value $VersionLevel -PropertyType String -Force | Out-Null
    "-----------------------------------------------------------------------------------------" | Out-File -Append $installLogFile
    $global:PsqlInfo=Get-ItemProperty $registryPath"\Products\"$InstallInfoProdType"\InstallInfo"
    $PsqlInfo | Format-List | Out-File -Append $installLogFile
    "-----------------------------------------------------------------------------------------" | Out-File -Append $installLogFile
    ## ELS Key
    if (Test-Path $registryPath"\ELS") {
        $msg.skipElsKey -f "$registryPath\ELS" | Tee-Object -FilePath $installLogFile -Append
    } else {
        $msg.addElsKey -f "$registryPath\ELS" | Tee-Object -FilePath $installLogFile -Append
        New-Item -Path $registryPath"\ELS" -Force | Out-Null
    }
    $msg.addElsValues -f $installLogFile | Tee-Object -FilePath $installLogFile -Append
    New-ItemProperty -Path $registryPath"\ELS" -Name "elsh" -Value $psqlElsURL -PropertyType String -Force | Out-Null
    New-ItemProperty -Path $registryPath"\ELS" -Name "elshid" -Value $psqlElsServerID -PropertyType DWORD -Force | Out-Null
    New-ItemProperty -Path $registryPath"\ELS" -Name "logfile" -Value $PsqlDataDir"logs\els.log" -PropertyType String -Force | Out-Null
    "-----------------------------------------------------------------------------------------" | Out-File -Append $installLogFile
    Get-Item $registryPath"\ELS" | Out-File -Append $installLogFile
    "-----------------------------------------------------------------------------------------" | Out-File -Append $installLogFile
    "" | Tee-Object -FilePath $installLogFile -Append
}
function RegisterComInterfaces {
    ##########################
    # Checks to see which PSQL COM interface library files are installed and then registers
    # each file using the regsvr32.exe utility.
    ##########################
    $msg.doRegisterCom  | Tee-Object -FilePath $installLogFile -Append
    $nFound=0;
    $nRegistered=0;
    if ($Is64Bit) {$filePath=$PsqlInstDir64} else {$filePath=$PsqlInstDir32}
    foreach ($fileToRegister in $ComFileList) {
        $fullFilePath=$filePath+"bin\"+$fileToRegister
        if (Test-Path $fullFilePath) {
            $nFound++
            & regsvr32.exe -s "$fullFilePath"| Out-String -Stream
            switch ($LASTEXITCODE) {
                0       {$msg.passRegFile -f $fullFilePath | Out-File -Append $installLogFile; $nRegistered++;break}
                default {$msg.warnRegFile -f $fullFilePath | Tee-Object -FilePath $installLogFile -Append; break}
            }
        }
    }
    "Successfully self-registered $nRegistered out of $nFound files. Refer to $installLogFile for more information." | Tee-Object -FilePath $installLogFile -Append
    "" | Tee-Object -FilePath $installLogFile -Append
}
function CreateServiceIfNotExist {
    ##########################
    # Wrapper for New-Service commandlet that creates a new service only after
    # first checking to verify the specified binary exists and the service
    # doesn't already exist.
    # Assumes arguments are specified in the following order:
    #   $args[0]: Path to binary.
    #   $args[1]: Service name
    #   $args[2]: Service display name
    #   $args[3]: Service description
    #   $args[4]: service arguments (if any)
    ##########################
    $svcPath=[string]$args[0]+[string]$args[4]
    $svcName=[string]$args[1]
    $svcDisplayName=[string]$args[2]
    $svcDesc=[string]$args[3]
    if (Test-Path $args[0]) {
        if (get-Service -Name $svcName -errorAction Ignore) {
            $msg.skipAddSvc -f $svcName | Tee-Object -FilePath $installLogFile -Append
        } else {
            New-Service -BinaryPathName $svcPath -Name $svcName -DisplayName $svcDisplayName -Description $svcDesc  -StartupType Automatic  -ErrorAction Ignore | Out-Null
            if ($?) {
               $msg.passAddSvc -f $svcDisplayName | Tee-Object -FilePath $installLogFile -Append
            } else {
                $msg.errAddSvc -f $svcDisplayName | Tee-Object -FilePath $installLogFile -Append
                "" | Tee-Object -FilePath $installLogFile -Append
                exit -1
            }
        }
    } else {
        $msg.warnNoSvcFile -f [string]$args[0], $svcDisplayName | Tee-Object -FilePath $installLogFile -Append
    }
}
function CreatePsqlServices {
    ##########################
    # Specifies the service parameters for each product type and then passes
    # them to CreateServiceIfNotExist to verify requires, create the service
    # if it doesn't already exist, and log transactions.
    ##########################
    if ( $isClient) {
        # -- No services are created for client installs.
        return
    }
    $msg.doAddSvc | Tee-Object -FilePath $installLogFile -Append
    if ($Is64Bit) {
        if ($isServer -or $isVxServer) {
            # 64-bit Server Engine
            $svcPath=$PsqlInstDir64 + "bin\ntbtrv64.exe"
            $svcName="Pervasive.SQL (transactional)"
            $svcDisplayName="Actian PSQL Transactional Engine"
            $svcDesc=$msg.svcDesc -f $svcDisplayName
            $svcArgs=""
            CreateServiceIfNotExist $svcPath $svcName $svcDisplayName $svcDesc $svcArgs

            $svcPath=$PsqlInstDir64 + "bin\w64sqlmgr.exe"
            $svcName="Pervasive.SQL (relational)"
            $svcDisplayName="Actian PSQL Relational Engine"
            $svcDesc=$msg.svcDesc -f $svcDisplayName
            $svcArgs=""
            CreateServiceIfNotExist $svcPath $svcName $svcDisplayName $svcDesc $svcArgs
        }
        if ($isReporting) {
            # 64-bit Reporting Engine
            $svcPath=$PsqlInstDir64 + "bin\w64dbsmgr.exe"
            $svcName="psqlRepEng"
            $svcDisplayName="Actian PSQL Reporting Engine"
            $svcDesc=$msg.svcDesc -f $svcDisplayName
            $svcArgs=" -service -srde"
            CreateServiceIfNotExist $svcPath $svcName $svcDisplayName $svcDesc $svcArgs
        }
    } else {
        if ($isServer -or $isVxServer) {
            # 32-bit Server Engine
            $svcPath=$PsqlInstDir32 + "bin\ntbtrv.exe"
            $svcName="Pervasive.SQL (transactional)"
            $svcDisplayName="Actian PSQL Transactional Engine"
            $svcDesc=$msg.svcDesc -f $svcDisplayName
            $svcArgs=""
            CreateServiceIfNotExist $svcPath $svcName $svcDisplayName $svcDesc $svcArgs

            $svcPath=$PsqlInstDir32 + "bin\w3sqlmgr.exe"
            $svcName="Pervasive.SQL (relational)"
            $svcDisplayName="Actian PSQL Relational Engine"
            $svcDesc=$msg.svcDesc -f $svcDisplayName
            $svcArgs=""
            CreateServiceIfNotExist $svcPath $svcName $svcDisplayName $svcDesc $svcArgs
        }
        if ($isWorkgroup) {
            # 32-bit Workgroup Engine
            $svcPath=$PsqlInstDir32 + "bin\w3dbsmgr.exe"
            $svcName="psqlWGE"
            $svcDisplayName="Actian PSQL Workgroup Engine"
            $svcDesc=$msg.svcDesc -f $svcDisplayName
            $svcArgs=" -service -srde"
            CreateServiceIfNotExist $svcPath $svcName $svcDisplayName $svcDesc $svcArgs
        }
    }
    "" | Tee-Object -FilePath $installLogFile -Append
}
function AddFirewallRuleIfNotExist {
    ##########################
    # Create the specified firewall rule only if the specified binary exists and
    # the specified rule name does not already exist.  The specified rule is
    # created but is NOT ENABLED by default.
    # Assumes arguments are specified in the following order:
    #   $args[0]: Path to binary
    #   $args[1]: rule name
    ##########################
    $rulePath=$args[0]
    $ruleName=$args[1]
    if (Test-Path $rulePath) {
        Try {
            # Use "-ErrorAction Stop" to trigger failure when Get-NetFirewallRule cmdlet fails.
            Get-NetFirewallRule -DisplayName $ruleName -ErrorAction Stop | Out-Null
            $msg.skipAddRule -f $ruleName | Tee-Object -FilePath $installLogFile -Append
        }
        Catch {
            # -- Rule doesn't exist, create it. Always disable the rule for all firewall profiles.
            Try {
                New-NetFirewallRule -DisplayName $ruleName -Direction Inbound -Program $rulePath -Action Allow -Profile Any -Enabled False | Out-File -Append $installLogFile
                if ($?) { $msg.passAddRule -f $ruleName | Tee-Object -FilePath $installLogFile -Append }
            }
            Catch {$msg.warnAddRuleFailed -f $ruleName | Tee-Object -FilePath $installLogFile -Append}
        }

    } else {$msg.warnNoRuleFile -f $rulePath, $ruleName | Tee-Object -FilePath $installLogFile -Append}
}
function AddPsqlFirewallRules {
    ##########################
    # If the NetSecurity module is present this function defines firewall rule
    # properties based on product type and then calls AddFirewallRuleIfNotExist
    # to create the specified rule(s) for the current product type.
    ##########################
    if ( $isClient) {
        # -- No firewall rules are created for client installs.
        return
    }
    $msg.doAddFWRules | Tee-Object -FilePath $installLogFile -Append
    $moduleNetSecurity=$false
    Get-Command -CommandType Function | ForEach-Object -Process {if ($_.Source -eq "NetSecurity") {$moduleNetSecurity=$true} }
    if ($moduleNetSecurity) {
        Import-Module -Name NetSecurity  # -- Firewall rule cmdlets.
    } else {
        # -- The NetSecurity Module is not installed, warn user and exit function.
        $msg.warnNoNetSecCreate | Tee-Object -FilePath $installLogFile -Append
        "" | Tee-Object -FilePath $installLogFile -Append
        return
    }
    if ($Is64Bit) {
      if ($isServer -or $isVxServer) {
          $fwRulePath=$PsqlInstDir64 + "bin\w64sqlmgr.exe"
          $fwRuleName="Actian PSQL Relational Engine"
          AddFirewallRuleIfNotExist $fwRulePath $fwRuleName
          $fwRulePath=$PsqlInstDir64 + "bin\ntbtrv64.exe"
          $fwRuleName="Actian PSQL Transactional Engine"
          AddFirewallRuleIfNotExist $fwRulePath $fwRuleName
          $fwRulePath=$PsqlInstDir64 + "bin\ntdbsmgr64.exe"
          $fwRuleName="Actian PSQL Database Engine"
          AddFirewallRuleIfNotExist $fwRulePath $fwRuleName
      }
      if ($isReporting) {
          $fwRulePath=$PsqlInstDir64 + "bin\w64dbsmgr.exe"
          $fwRuleName="Actian PSQL Reporting Engine"
          AddFirewallRuleIfNotExist $fwRulePath $fwRuleName
      }
    } else {
      if ($isServer -or $isVxServer) {
          $fwRulePath=$PsqlInstDir32 + "bin\w3sqlmgr.exe"
          $fwRuleName="Actian PSQL Relational Engine"
          AddFirewallRuleIfNotExist $fwRulePath $fwRuleName
          $fwRulePath=$PsqlInstDir32 + "bin\ntbtrv.exe"
          $fwRuleName="Actian PSQL Transactional Engine"
          AddFirewallRuleIfNotExist $fwRulePath $fwRuleName
          $fwRulePath=$PsqlInstDir32 + "bin\ntdbsmgr.exe"
          $fwRuleName="Actian PSQL Database Engine"
          AddFirewallRuleIfNotExist $fwRulePath $fwRuleName
      }
      if ($isWorkgroup) {
          $fwRulePath=$PsqlInstDir32 + "bin\w3dbsmgr.exe"
          $fwRuleName="Actian PSQL Workgroup Engine"
          AddFirewallRuleIfNotExist $fwRulePath $fwRuleName
      }
    }
    "" | Tee-Object -FilePath $installLogFile -Append
}
function StartPsqlServices {
    ##########################
    # Start any service with a display name beginning with the string
    # identified by $CmnSvcDispName.
    ##########################
    $msg.doStartPsqlSvcs | Tee-Object -FilePath $installLogFile -Append
    $foundSvc=$false
    get-service | where {$_.DisplayName -Like $CmnSvcDispName+"*"} | Sort-Object -Descending |
        ForEach-Object -Process {
            $foundSvc=$true
            if ($_.Status -ne "Running") {
                $svcDisplayName=$_.DisplayName
                $svcName=$_.Name
                Start-Service -Name $svcName -ErrorAction Ignore| Out-Null
                if ($?) {
                    $msg.passStartSvcs -f $svcDisplayName | Tee-Object -FilePath $installLogFile -Append
                } else {
                    $msg.errStartSvcs -f $svcDisplayName | Tee-Object -FilePath $installLogFile -Append
                    "" | Tee-Object -FilePath $installLogFile -Append
                    exit -1
                }
                # Wait 5 seconds.
                Start-Sleep -Milliseconds 5000
            } else {
                $msg.skipStartSvcs -f $svcDisplayName | Tee-Object -FilePath $installLogFile -Append
            }
        }
    if (!$foundSvc) {$msg.notFoundSvcs -f $CmnSvcDispName | Tee-Object -FilePath $installLogFile -Append}
    "" | Tee-Object -FilePath $installLogFile -Append
}
function CreatePsqlDatabases {
    ##########################
    # Runs the command line setupdbcli.exe utility to create the system databases, bulk load data
    # into them, and execute SQL statements to configure them as determined by the settings
    # specified in the configuraiton file being processed.
    #
    # PREREQ: The PSQL Service(s) must be started prior to calling setupdbcli.exe.
    ##########################
    if ( $isReporting -or $isClient) {
        # -- No databases are created during install for client or reporting engine.
        $msg.skipCreateDbs | Tee-Object -FilePath $installLogFile -Append
        "" | Tee-Object -FilePath $installLogFile -Append
        return
    }

    # -- Continue (Server, Vx, or Workgroup only).
    $msg.doCreateDbs | Tee-Object -FilePath $installLogFile -Append

    # -- Verify setupdb configuration file exists.
    $setupdbLog=$PsqlDataDir+"Logs\PSQL_Setupdb.log"
    if (Test-Path $setupdbCfgInstall) {
        $setupdb_args="-log",$setupdbLog,"-dir",$PsqlDataDir,$setupdbCfgInstall
    } else {
        $msg.noSetupDbCfgCreate -f $setupdbCfgInstall | Tee-Object -FilePath $installLogFile -Append
    }
    if (Test-Path $PsqlDataDir"DBNAMES.CFG") {
        $msg.foundDbnames -f $PsqlDataDir | Tee-Object -FilePath $installLogFile -Append
    } else {
        $msg.notFoundDbnames -f $PsqlDataDir | Tee-Object -FilePath $installLogFile -Append
    }

    # -- Execute setupdbcli.exe.
    if ($Is64Bit) {
        if ($isServer -or $isVxServer) {
            if (Test-Path $PsqlInstDir64"bin\setupdbcli.exe") {
                # -- No return code from setupdbcli.exe so just log command line.
                $msg.runningSetupdb -f [string]$setupdb_args, $setupdbLog | Tee-Object -FilePath $installLogFile -Append
                Try {
                    Start-Process -FilePath $PsqlInstDir64"bin\setupdbcli.exe" -ArgumentList $setupdb_args -WorkingDirectory $PsqlInstDir64"bin" -NoNewWindow -Wait
                    if ($?) { $msg.passSetupdb -f $PsqlInstDir64, $setupdbLog | Tee-Object -FilePath $installLogFile -Append }
                }
                Catch  { $msg.failSetupdb -f $PsqlInstDir64 | Tee-Object -FilePath $installLogFile -Append }
            } else {
                $msg.noSetupDb -f $PsqlInstDir64 | Tee-Object -FilePath $installLogFile -Append
            }
        }
    } else {
        if (Test-Path $PsqlInstDir32"bin\setupdbcli.exe") {
            # -- No return code from setupdbcli.exe so just log command line.
            $msg.runningSetupdb -f [string]$setupdb_args, $setupdbLog | Tee-Object -FilePath $installLogFile -Append
            Try {
                Start-Process -FilePath $PsqlInstDir32"bin\setupdbcli.exe" -ArgumentList $setupdb_args -WorkingDirectory $PsqlInstDir32"bin" -NoNewWindow -Wait
                if ($?) { $msg.passSetupdb -f $PsqlInstDir32, $setupdbLog | Tee-Object -FilePath $installLogFile -Append }
            }
            Catch  {  $msg.failSetupdb -f $PsqlInstDir32 | Tee-Object -FilePath $installLogFile -Append }
        } else {
            $msg.noSetupDb -f $PsqlInstDir32 | Tee-Object -FilePath $installLogFile -Append
        }
    }
    "" | Tee-Object -FilePath $installLogFile -Append
}
function ActivatePsqlTempLicense {
    ##########################
    # Calls the appropriate command line license administrator utility to add the temp
    # license key defined in $PsqlInstallCfg and stored in $PVSW_PSQL_LICENSE_KEY per
    # the previous call to the ReadFromConfig function.
    ##########################
    if ( $isReporting -or $isClient -or $updating) {
        # -- No license required for client or reporting engine or if updating.
        return
    }

    # -- Continue (Server, Vx, or Workgroup only).
    $msg.doAddTempLic | Tee-Object -FilePath $installLogFile -Append
    $result=$false
    if ($Is64Bit) {
        if (Test-Path $PsqlInstDir64"bin\clilcadm.exe") {
            $result=& clilcadm.exe -a $PVSW_PSQL_LICENSE_KEY| Out-String -Stream
        } else {
            $msg.notFoundClilcadm -f $PsqlInstDir64 | Tee-Object -FilePath $installLogFile -Append
        }
    } else {
        if (Test-Path $PsqlInstDir32"bin\clilcadm.exe") {
            $result=& clilcadm.exe -a $PVSW_PSQL_LICENSE_KEY| Out-String -Stream
        } else {
            $msg.notFoundClilcadm -f $PsqlInstDir32 | Tee-Object -FilePath $installLogFile -Append
        }
    }

    # -- Interpret results.
    if ($result) {
        switch ($lastExitCode) {
            0       {$msg.passAddTempLic -f $PVSW_PSQL_LICENSE_KEY | Tee-Object -FilePath $installLogFile -Append; break}
            default {$msg.warnAddTempLic -f $result[4], $result[5], $result[6] | Tee-Object -FilePath $installLogFile -Append; break}
        }
    }
    "" | Tee-Object -FilePath $installLogFile -Append
}
function InvokeECollect {
    ##########################
    # Wrapper function for making DTI call pvCollectLocal() to collect the
    # PSQL product information during install or update.  This call requires
    # the appropriate ELS library is installed so we skip this step if the
    # library is not found.
    #
    # API Invoked
    #   32-bit systems: clientlm.dll (requires w3els.dll)
    #   64-bit systems: w64clientlm.dll (requires w64els.dll)
    #     install - pvCollectLocal(COLLECT_ON_INSTALL)
    #     update  - pvCollectLocal(COLLECT_ON_PATCH)
    ##########################

    # Skip eCollect if the appropriate ELS library is not present or if this is
    # a client or Reporting Engine install (pvCollectLocal only supported during engine installs).
    if ($Is64Bit) { $elsLib=$PsqlInstDir64 + "bin\w64els.dll"} else {$elsLib=$PsqlInstDir32 + "bin\w3els.dll"}
    if (!(Test-Path $elsLib) -or $isClient -or $isReporting ) {
        $msg.skipECollect | Tee-Object -FilePath $installLogFile -Append
        "" | Tee-Object -FilePath $installLogFile -Append
        return
    }

    # Not sure why but the ELS libraries are requiring the PSQL services to be restarted
    # on install after the system databases are created during CreatePsqlDatabases().
    StopPsqlServices
    StartPsqlServices

    $msg.doECollect | Tee-Object -FilePath $installLogFile -Append
    if ($updating) {[int]$eventId=120} else {[int]$eventId=100}
    if ($Is64Bit) {$dllName="w64clientlm.dll"} else {$dllName="clientlm.dll"}

    # Create the eCollect ScriptBlock. We need to invoke the ScriptBlock as a
    # background task because PowerShell has no way of unloading a DLL after the
    # Add-Type cmdlet has been used, resulting in w64/clientlm.dll and all
    # dependent libraries being locked.
    # Note the two lines in the 'here string' that identify the ELS library
    # signature can't have any leading spaces.
    $apiScriptBlock= {
        param ([string]$dllFileName="clientlm.dll", [int]$eCollectEventId=100)
        $ApiSignature = @"
[DllImport("$dllFileName")]
public static extern int pvCollectLocal(int nEventId);
"@
        Try {
            $pvCollectLocal= Add-Type -MemberDefinition $ApiSignature -Name "Win32pvCollectLocal" -Namespace Win32Functions -PassThru
            $pvCollectLocal::pvCollectLocal($eCollectEventId)
        }
        Catch {[string]$error[0]}
    }

    # Execute the scriptblock with the appropriate library name and event ID.
    Try {
        Start-Job -Name eCollect -ScriptBlock $apiScriptBlock -args $dllName,$eventId | Out-Null
        Wait-Job -Name eCollect | Out-Null
        $result=Receive-Job -Name eCollect
        if ($result.GetType().ToString() -eq "System.Int32") {
            # ScriptBlock returned int.
            if ($result -eq 0) {
                $msg.passECollect | Tee-Object -FilePath $installLogFile -Append
            } else {
                $msg.warnECollect -f "pvCollectLocal call",$result | Tee-Object -FilePath $installLogFile -Append
            }
        } else {
           # ScriptBlock returned non-int.
           $msg.warnECollect -f "eCollect ScriptBlock",$result | Tee-Object -FilePath $installLogFile -Append
        }
    }
    Catch {$msg.warnECollect -f "eCollect ScriptBlock",$result | Tee-Object -FilePath $installLogFile -Append}
    "" | Tee-Object -FilePath $installLogFile -Append
}
function UpdateRoyaltyRecord {
    ##########################
    # Wrapper function for making DTI call PvValidateLicenses() to update the
    # royalty record (also creates the 'Product Data' registry key).
    #
    # API Invoked
    #   32-bit systems: w3dbav90.dll::PvValidateLicenses(0xFFFFFFFF)
    #   64-bit systems: w64dba.dll::PvValidateLicenses(0xFFFFFFFF)
    ##########################

    # Skip if this is a client or reporting engine install (PvValidateLicenses only supported during engine installs).
    if ( $isClient  -or $isReporting) {return}

    $msg.doUpdateRoyaltyRec | Tee-Object -FilePath $installLogFile -Append
    if ($Is64Bit) {$dllName="w64dba.dll"} else {$dllName="w3dbav90.dll"}

    # Create the ScriptBlock to update the royalty record. We need to invoke
    # the ScriptBlock as a background task because PowerShell has no way of
    # unloading a DLL after the Add-Type cmdlet has been used, resulting in
    # library file and all dependent libraries being locked.
    # Note the two lines in the 'here string' that identify the DTI library
    # signature can't have any leading spaces.
    $apiScriptBlock= {
        param ([string]$dllFileName="w3dbav90.dll")
        $ApiSignature = @"
[DllImport("$dllFileName")]
public static extern int PvValidateLicenses(int hConnection);
"@
        Try {
            [long]$hConnection=0xFFFFFFFF
            $PvValidateLicenses= Add-Type -MemberDefinition $ApiSignature -Name "Win32pvCollectLocal" -Namespace Win32Functions -PassThru
            $PvValidateLicenses::PvValidateLicenses($hConnection)
        }
        Catch {[string]$error[0]}
    }

    # Execute the scriptblock with the appropriate library name and event ID.
    Try {
        Start-Job -Name eCollect -ScriptBlock $apiScriptBlock -args $dllName | Out-Null
        Wait-Job -Name eCollect | Out-Null
        $result=Receive-Job -Name eCollect
        if ($result.GetType().ToString() -eq "System.Int32") {
            # ScriptBlock returned int.
            if ($result -eq 0) {
                $msg.passUpdateRoyaltyRec | Tee-Object -FilePath $installLogFile -Append
            } else {
                $msg.warnUpdateRoyaltyRec -f "PvValidateLicenses call",$result | Tee-Object -FilePath $installLogFile -Append
            }
        } else {
           # ScriptBlock returned non-int.
           $msg.warnUpdateRoyaltyRec -f "Update royalty record ScriptBlock",$result | Tee-Object -FilePath $installLogFile -Append
        }
    }
    Catch {$msg.warnUpdateRoyaltyRec -f "eCollect ScriptBlock",$result | Tee-Object -FilePath $installLogFile -Append}
    "" | Tee-Object -FilePath $installLogFile -Append
}
function LoadPerfMonCounterSets {
    ##########################
    # Wrapper for Lodctr.exe, called to load the PSQL counter sets for
    # Windows Performance Monitor.  This is only run during Server Engine,
    # Vx Server, or Reporting Engine installs and only if lodctr.exe is
    # found.
    #
    # Executes:
    #   "$env:windir\system32\lodctr.exe" {path}\{XML config file}
    ##########################
    # Skip if WGE or Client install.
    if ( $isWorkgroup -or $isClient) {
        # -- Windows Performance Monitor counter sets are not installed for WGE or Client.
        return
    }
    # Skip if system doesn't support it (no lodctr.exe).
    $loader="$env:windir\system32\lodctr.exe"
    if (!(Test-Path $loader)) {return}

    # Continue (lodctr.exe found and SE, Vx, or Rpt engine).
    $msg.doLoadPerfMonCtr | Tee-Object -FilePath $installLogFile -Append
    ## Set the configuration file.
    if ($isServer -or $isVxServer) {
        if ($Is64Bit) {
            $perfctrcfg=$PsqlInstDir64 + "bin\psqlperfctrSE64.xml"
        } else {
            $perfctrcfg=$PsqlInstDir32 + "bin\psqlperfctrSE32.xml"
        }
    } else {
        if ($isReporting) { $perfctrcfg=$PsqlInstDir64 + "bin\psqlperfctrRE64.xml" }
    }
    if (Test-Path $perfctrcfg) {
        Try {
            $result=& LodCtr.exe /m:"$perfctrcfg" | Out-String -Stream
            if ($LASTEXITCODE -eq 0) {
                $msg.passLoadPerfMonCtr | Tee-Object -FilePath $installLogFile -Append
            } else {
                $msg.warnLoadPerfMonCtr -f $perfctrcfg, [string]$LASTEXITCODE, [string]$result | Tee-Object -FilePath $installLogFile -Append
            }
        }
        Catch { $msg.warnLoadPerfMonCtr -f $perfctrcfg, [string]$LASTEXITCODE, [string]$result | Tee-Object -FilePath $installLogFile -Append }
    } else {
        $msg.warnNotFoundPerfMonCtrCfg | Tee-Object -FilePath $installLogFile -Append
    }
    "" | Tee-Object -FilePath $installLogFile -Append
}
function EndLog {
    ##########################
    # Close $installLogFile with date/time script completed.
    ##########################
    $msg.infoEnd -f [DateTime]::Parse((Get-Date).ToString()) | Out-File -Append $installLogFile
    ##########################
    "#############################" | Tee-Object -FilePath $installLogFile -Append
    "" | Out-File -Append $installLogFile
    $msg.completeInstallMsg -f $installLogFile
    "" | Tee-Object -FilePath $installLogFile -Append
}

Try {. Main}
Catch {throw $msg.errUnknown; exit -1}
# SIG # Begin signature block
# MIIcMwYJKoZIhvcNAQcCoIIcJDCCHCACAQExCzAJBgUrDgMCGgUAMGkGCisGAQQB
# gjcCAQSgWzBZMDQGCisGAQQBgjcCAR4wJgIDAQAABBAfzDtgWUsITrck0sYpfvNR
# AgEAAgEAAgEAAgEAAgEAMCEwCQYFKw4DAhoFAAQUtH1UcpfjUWKFVVoYVzT+RqCf
# ogKgghcYMIID7jCCA1egAwIBAgIQfpPr+3zGTlnqS5p31Ab8OzANBgkqhkiG9w0B
# AQUFADCBizELMAkGA1UEBhMCWkExFTATBgNVBAgTDFdlc3Rlcm4gQ2FwZTEUMBIG
# A1UEBxMLRHVyYmFudmlsbGUxDzANBgNVBAoTBlRoYXd0ZTEdMBsGA1UECxMUVGhh
# d3RlIENlcnRpZmljYXRpb24xHzAdBgNVBAMTFlRoYXd0ZSBUaW1lc3RhbXBpbmcg
# Q0EwHhcNMTIxMjIxMDAwMDAwWhcNMjAxMjMwMjM1OTU5WjBeMQswCQYDVQQGEwJV
# UzEdMBsGA1UEChMUU3ltYW50ZWMgQ29ycG9yYXRpb24xMDAuBgNVBAMTJ1N5bWFu
# dGVjIFRpbWUgU3RhbXBpbmcgU2VydmljZXMgQ0EgLSBHMjCCASIwDQYJKoZIhvcN
# AQEBBQADggEPADCCAQoCggEBALGss0lUS5ccEgrYJXmRIlcqb9y4JsRDc2vCvy5Q
# WvsUwnaOQwElQ7Sh4kX06Ld7w3TMIte0lAAC903tv7S3RCRrzV9FO9FEzkMScxeC
# i2m0K8uZHqxyGyZNcR+xMd37UWECU6aq9UksBXhFpS+JzueZ5/6M4lc/PcaS3Er4
# ezPkeQr78HWIQZz/xQNRmarXbJ+TaYdlKYOFwmAUxMjJOxTawIHwHw103pIiq8r3
# +3R8J+b3Sht/p8OeLa6K6qbmqicWfWH3mHERvOJQoUvlXfrlDqcsn6plINPYlujI
# fKVOSET/GeJEB5IL12iEgF1qeGRFzWBGflTBE3zFefHJwXECAwEAAaOB+jCB9zAd
# BgNVHQ4EFgQUX5r1blzMzHSa1N197z/b7EyALt0wMgYIKwYBBQUHAQEEJjAkMCIG
# CCsGAQUFBzABhhZodHRwOi8vb2NzcC50aGF3dGUuY29tMBIGA1UdEwEB/wQIMAYB
# Af8CAQAwPwYDVR0fBDgwNjA0oDKgMIYuaHR0cDovL2NybC50aGF3dGUuY29tL1Ro
# YXd0ZVRpbWVzdGFtcGluZ0NBLmNybDATBgNVHSUEDDAKBggrBgEFBQcDCDAOBgNV
# HQ8BAf8EBAMCAQYwKAYDVR0RBCEwH6QdMBsxGTAXBgNVBAMTEFRpbWVTdGFtcC0y
# MDQ4LTEwDQYJKoZIhvcNAQEFBQADgYEAAwmbj3nvf1kwqu9otfrjCR27T4IGXTdf
# plKfFo3qHJIJRG71betYfDDo+WmNI3MLEm9Hqa45EfgqsZuwGsOO61mWAK3ODE2y
# 0DGmCFwqevzieh1XTKhlGOl5QGIllm7HxzdqgyEIjkHq3dlXPx13SYcqFgZepjhq
# IhKjURmDfrYwggR9MIIDZaADAgECAgMb5xUwDQYJKoZIhvcNAQELBQAwYzELMAkG
# A1UEBhMCVVMxITAfBgNVBAoTGFRoZSBHbyBEYWRkeSBHcm91cCwgSW5jLjExMC8G
# A1UECxMoR28gRGFkZHkgQ2xhc3MgMiBDZXJ0aWZpY2F0aW9uIEF1dGhvcml0eTAe
# Fw0xNDAxMDEwNzAwMDBaFw0zMTA1MzAwNzAwMDBaMIGDMQswCQYDVQQGEwJVUzEQ
# MA4GA1UECBMHQXJpem9uYTETMBEGA1UEBxMKU2NvdHRzZGFsZTEaMBgGA1UEChMR
# R29EYWRkeS5jb20sIEluYy4xMTAvBgNVBAMTKEdvIERhZGR5IFJvb3QgQ2VydGlm
# aWNhdGUgQXV0aG9yaXR5IC0gRzIwggEiMA0GCSqGSIb3DQEBAQUAA4IBDwAwggEK
# AoIBAQC/cWII8fpZNPcbyRij94BJWOkigxOmxSBDATuE8eaFSZ8n6vaEG06gtNtw
# mMcyAbEFPgdO7vT6Ty9ZMCLnqxlWa+KAB/zzFnWAOVF75fk1tnROqY2CE+S2P6kD
# g/qivooVan/eC8O2GRQFyurDqASUO0Z8Mg3zAGYiyI1pbTaMERi307IcYLQ4+gKM
# ztPdRgfeCj7rXXzIfPuwK1OkkmJpUSUFYRpEgYwsqUOWI9+sOoGaDinFHKnpXR62
# np4wCjnO8YiA+0tdzDLshWJDJTQCVicBkbQ7cCo/brHonIgBfZ/U+dtTbWCdvyzn
# WKu4X0b8zsQbAzwJ60kxXGlGs+BHAgMBAAGjggEXMIIBEzAPBgNVHRMBAf8EBTAD
# AQH/MA4GA1UdDwEB/wQEAwIBBjAdBgNVHQ4EFgQUOpqFBxBnKLbv9r0FQW4gwZTa
# D94wHwYDVR0jBBgwFoAU0sSw0pHUTBFxs2HLPaH+3ahq1OMwNAYIKwYBBQUHAQEE
# KDAmMCQGCCsGAQUFBzABhhhodHRwOi8vb2NzcC5nb2RhZGR5LmNvbS8wMgYDVR0f
# BCswKTAnoCWgI4YhaHR0cDovL2NybC5nb2RhZGR5LmNvbS9nZHJvb3QuY3JsMEYG
# A1UdIAQ/MD0wOwYEVR0gADAzMDEGCCsGAQUFBwIBFiVodHRwczovL2NlcnRzLmdv
# ZGFkZHkuY29tL3JlcG9zaXRvcnkvMA0GCSqGSIb3DQEBCwUAA4IBAQBZC1O9koYR
# pyR77Vsxzx0fbHDFuG6+Trv2vpdQ4TB/uihcYpTC434z9/tCdoXblRyMIlh1CQyI
# ZWc5ChYJxaA4l6TFI5M/tBimAQZEkeOnaSe0WiV/Orcyzd2E/yo4KTOk3Weyhf6h
# iCAcUInI3Cr2QgM3TOaI39WvJPKxw9/MtezgmV63SVQgPJQYDMccUhhJpG3hs1gL
# ydjs2a4cMo4ocA3i/qYXnoQPvVdws1rpH6CGU7vvfP9pC+BIw7eTC8gKVMSsXRRn
# N2zKpS8xCDeqbm+MvJviV10kga+Xl5yErWysN0xm82GRESDkvjCfeqQpCbDhNF9k
# dxhAUd+MMKavMIIEozCCA4ugAwIBAgIQDs/0OMj+vzVuBNhqmBsaUDANBgkqhkiG
# 9w0BAQUFADBeMQswCQYDVQQGEwJVUzEdMBsGA1UEChMUU3ltYW50ZWMgQ29ycG9y
# YXRpb24xMDAuBgNVBAMTJ1N5bWFudGVjIFRpbWUgU3RhbXBpbmcgU2VydmljZXMg
# Q0EgLSBHMjAeFw0xMjEwMTgwMDAwMDBaFw0yMDEyMjkyMzU5NTlaMGIxCzAJBgNV
# BAYTAlVTMR0wGwYDVQQKExRTeW1hbnRlYyBDb3Jwb3JhdGlvbjE0MDIGA1UEAxMr
# U3ltYW50ZWMgVGltZSBTdGFtcGluZyBTZXJ2aWNlcyBTaWduZXIgLSBHNDCCASIw
# DQYJKoZIhvcNAQEBBQADggEPADCCAQoCggEBAKJjCzlEuLsjp0RJuw7/ofBhClOT
# sJjbrSwPSsVu/4Y8U1UPFc4EPyv9qZaW2b5heQtbyUyGduXgQ0sile7CK0PBn9ho
# tI5AT+6FOLkRxSPyZFjwFTJvTlehroikAtcqHs1L4d1j1ReJMluwXplaqJ0oUA4X
# 7pbbYTtFUR3PElYLkkf8q672Zj1HrHBy55LnX80QucSDZJQZvSWA4ejSIqXQugJ6
# oXeTW2XD7hd0vEGGKtwITIySjJEtnndEH2jWqHR32w5bMotWizO92WPISZ06xcXq
# MwvS8aMb9Iu+2bNXizveBKd6IrIkri7HcMW+ToMmCPsLvalPmQjhEChyqs0CAwEA
# AaOCAVcwggFTMAwGA1UdEwEB/wQCMAAwFgYDVR0lAQH/BAwwCgYIKwYBBQUHAwgw
# DgYDVR0PAQH/BAQDAgeAMHMGCCsGAQUFBwEBBGcwZTAqBggrBgEFBQcwAYYeaHR0
# cDovL3RzLW9jc3Aud3Muc3ltYW50ZWMuY29tMDcGCCsGAQUFBzAChitodHRwOi8v
# dHMtYWlhLndzLnN5bWFudGVjLmNvbS90c3MtY2EtZzIuY2VyMDwGA1UdHwQ1MDMw
# MaAvoC2GK2h0dHA6Ly90cy1jcmwud3Muc3ltYW50ZWMuY29tL3Rzcy1jYS1nMi5j
# cmwwKAYDVR0RBCEwH6QdMBsxGTAXBgNVBAMTEFRpbWVTdGFtcC0yMDQ4LTIwHQYD
# VR0OBBYEFEbGaaMOShQe1UzaUmMXP142vA3mMB8GA1UdIwQYMBaAFF+a9W5czMx0
# mtTdfe8/2+xMgC7dMA0GCSqGSIb3DQEBBQUAA4IBAQB4O7SRKgBM8I9iMDd4o4Qn
# B28Yst4l3KDUlAOqhk4ln5pAAxzdzuN5yyFoBtq2MrRtv/QsJmMz5ElkbQ3mw2cO
# 9wWkNWx8iRbG6bLfsundIMZxD82VdNy2XN69Nx9DeOZ4tc0oBCCjqvFLxIgpkQ6A
# 0RH83Vx2bk9eDkVGQW4NsOo4mrE62glxEPwcebSAe6xp9P2ctgwWK/F/Wwk9m1vi
# FsoTgW0ALjgNqCmPLOGy9FqpAa8VnCwvSRvbIrvD/niUUcOGsYKIXfA9tFGheTMr
# Lnu53CAJE3Hrahlbz+ilMFcsiUk/uc9/yb8+ImhjU5q9aXSsxR08f5Lgw7wc2AR1
# MIIE0DCCA7igAwIBAgIBBzANBgkqhkiG9w0BAQsFADCBgzELMAkGA1UEBhMCVVMx
# EDAOBgNVBAgTB0FyaXpvbmExEzARBgNVBAcTClNjb3R0c2RhbGUxGjAYBgNVBAoT
# EUdvRGFkZHkuY29tLCBJbmMuMTEwLwYDVQQDEyhHbyBEYWRkeSBSb290IENlcnRp
# ZmljYXRlIEF1dGhvcml0eSAtIEcyMB4XDTExMDUwMzA3MDAwMFoXDTMxMDUwMzA3
# MDAwMFowgbQxCzAJBgNVBAYTAlVTMRAwDgYDVQQIEwdBcml6b25hMRMwEQYDVQQH
# EwpTY290dHNkYWxlMRowGAYDVQQKExFHb0RhZGR5LmNvbSwgSW5jLjEtMCsGA1UE
# CxMkaHR0cDovL2NlcnRzLmdvZGFkZHkuY29tL3JlcG9zaXRvcnkvMTMwMQYDVQQD
# EypHbyBEYWRkeSBTZWN1cmUgQ2VydGlmaWNhdGUgQXV0aG9yaXR5IC0gRzIwggEi
# MA0GCSqGSIb3DQEBAQUAA4IBDwAwggEKAoIBAQC54MsQ1K92vdSTYuswZLiBCGzD
# BNliF44v/z5lz4/OYuY8UhzaFkVLVat4a2ODYpDOD2lsmcgaFItMzEUz6ojcnqOv
# K/6AYZ15V8TPLvQ/MDxdR/yaFrzDN5ZBUY4RS1T4KL7QjL7wMDge87Am+GZHY23e
# cSZHjzhHU9FGHbTj3ADqRay9vHHZqm8A29vNMDp5T19MR/gd71vCxJ1gO7GyQ5HY
# pDNO6rPWJ0+tJYqlxvTV0KaudAVkV4i1RFXULSo6Pvi4vekyCgKUZMQWOlDxSq7n
# eTOvDCAHf+jfBDnCaQJsY1L6d8EbyHSHyLmTGFBUNUtpTrw700kuH9zB0lL7AgMB
# AAGjggEaMIIBFjAPBgNVHRMBAf8EBTADAQH/MA4GA1UdDwEB/wQEAwIBBjAdBgNV
# HQ4EFgQUQMK9J47MNIMwojPX+2yz8LQsgM4wHwYDVR0jBBgwFoAUOpqFBxBnKLbv
# 9r0FQW4gwZTaD94wNAYIKwYBBQUHAQEEKDAmMCQGCCsGAQUFBzABhhhodHRwOi8v
# b2NzcC5nb2RhZGR5LmNvbS8wNQYDVR0fBC4wLDAqoCigJoYkaHR0cDovL2NybC5n
# b2RhZGR5LmNvbS9nZHJvb3QtZzIuY3JsMEYGA1UdIAQ/MD0wOwYEVR0gADAzMDEG
# CCsGAQUFBwIBFiVodHRwczovL2NlcnRzLmdvZGFkZHkuY29tL3JlcG9zaXRvcnkv
# MA0GCSqGSIb3DQEBCwUAA4IBAQAIfmyTEMg4uJapkEv/oV9PBO9sPpyIBslQj6Zz
# 91cxG7685C/b+LrTW+C05+Z5Yg4MotdqY3MxtfWoSKQ7CC2iXZDXtHwlTxFWMMS2
# RJ17LJ3lXubvDGGqv+QqG+6EnriDfcFDzkSnE3ANkR/0yBOtg2DZ2HKocyQetawi
# DsoXiWJYRBuriSUBAA/NxBti21G00w9RKpv0vHP8ds42pM3Z2Czqrpv1KrKQ0U11
# GIo/ikGQI31bS/6kA1ibRrLDYGCD+H1QQc7CoZDDu+8CL9IVVO5EFdkKrqeKM+2x
# LXY2JtwE65/3YR8V3Idv7kaWKK2hJn0KCacuBKONvPi8BDABMIIFJjCCBA6gAwIB
# AgIICsxCB6AJxjswDQYJKoZIhvcNAQELBQAwgbQxCzAJBgNVBAYTAlVTMRAwDgYD
# VQQIEwdBcml6b25hMRMwEQYDVQQHEwpTY290dHNkYWxlMRowGAYDVQQKExFHb0Rh
# ZGR5LmNvbSwgSW5jLjEtMCsGA1UECxMkaHR0cDovL2NlcnRzLmdvZGFkZHkuY29t
# L3JlcG9zaXRvcnkvMTMwMQYDVQQDEypHbyBEYWRkeSBTZWN1cmUgQ2VydGlmaWNh
# dGUgQXV0aG9yaXR5IC0gRzIwHhcNMTcwNDA1MjAwMjAwWhcNMjAwNDEwMTk1MTM5
# WjBoMQswCQYDVQQGEwJVUzEOMAwGA1UECBMFVGV4YXMxDzANBgNVBAcTBkF1c3Rp
# bjEbMBkGA1UEChMSQWN0aWFuIENvcnBvcmF0aW9uMRswGQYDVQQDExJBY3RpYW4g
# Q29ycG9yYXRpb24wggEiMA0GCSqGSIb3DQEBAQUAA4IBDwAwggEKAoIBAQDDl5bZ
# gsKcnWzrY8rYrsSmI4l9tC5Ipk1Dp/LHYZQ/1uwgzbiUQavCOKhJ6q+XMGHIOF6P
# N/PkfIfrf1XygR3bJy+f0u4HICioNg0GmfBgEUHVpZTkeRxCT+Dd1FeDDljp9J41
# AVkXh8SnzKvSGKjaUbp3DkH2Y5P24pHR9zhpx2hnqSF0PHOT0fGFcF4SAeYBTIT0
# nSmVPxjs2+NV1M81xtCLLLCF9ceV784DyY7MhoHI9Q7fHpxUJZi0nuHRzyUurofC
# QmJhOLd5vm7re/WUAHfhG3P7dIkm1mI9SHAWid2A/4vGWHb/bR5mwJTDxvPmgfzz
# EVmR8PHPNTsRXZYpAgMBAAGjggGFMIIBgTAMBgNVHRMBAf8EAjAAMBMGA1UdJQQM
# MAoGCCsGAQUFBwMDMA4GA1UdDwEB/wQEAwIHgDA1BgNVHR8ELjAsMCqgKKAmhiRo
# dHRwOi8vY3JsLmdvZGFkZHkuY29tL2dkaWcyczUtMi5jcmwwXQYDVR0gBFYwVDBI
# BgtghkgBhv1tAQcXAjA5MDcGCCsGAQUFBwIBFitodHRwOi8vY2VydGlmaWNhdGVz
# LmdvZGFkZHkuY29tL3JlcG9zaXRvcnkvMAgGBmeBDAEEATB2BggrBgEFBQcBAQRq
# MGgwJAYIKwYBBQUHMAGGGGh0dHA6Ly9vY3NwLmdvZGFkZHkuY29tLzBABggrBgEF
# BQcwAoY0aHR0cDovL2NlcnRpZmljYXRlcy5nb2RhZGR5LmNvbS9yZXBvc2l0b3J5
# L2dkaWcyLmNydDAfBgNVHSMEGDAWgBRAwr0njsw0gzCiM9f7bLPwtCyAzjAdBgNV
# HQ4EFgQUd28uU+LsLPArCXvwS49GX5IaKcUwDQYJKoZIhvcNAQELBQADggEBAKMz
# CNStFW8r/jaaSDXwbWnpmRjeGx3FZbmwH+r2gv2UQ5Se0ckdRm86CfF8AjGnFd/d
# pOSsLWUSWx5+gdqorB9+xY6KrB0br08CfRYvvw+ZYiRrnvLEJyD3hc13tHYpmDOS
# lamg3seFzk/rKe5g/ZXwDQXOiAMEkH5xBHS3BBILQarbiJ14CLM9eLDSeN8L/Gdy
# SvMX4nWy9UusiXP3ZXvuVPUlAOBvDUZaM1BPir0ZWY/m2FOW8bHxXzTG8VtdgY0y
# JIkaii7O3A1kkB9ofB47LQN/CuwcOOKQx0hNL3Nlq7zsFizWk6MO3ZmPNjjimkI6
# mT0NuoY6yd9VefcO95MxggSFMIIEgQIBATCBwTCBtDELMAkGA1UEBhMCVVMxEDAO
# BgNVBAgTB0FyaXpvbmExEzARBgNVBAcTClNjb3R0c2RhbGUxGjAYBgNVBAoTEUdv
# RGFkZHkuY29tLCBJbmMuMS0wKwYDVQQLEyRodHRwOi8vY2VydHMuZ29kYWRkeS5j
# b20vcmVwb3NpdG9yeS8xMzAxBgNVBAMTKkdvIERhZGR5IFNlY3VyZSBDZXJ0aWZp
# Y2F0ZSBBdXRob3JpdHkgLSBHMgIICsxCB6AJxjswCQYFKw4DAhoFAKCBijAZBgkq
# hkiG9w0BCQMxDAYKKwYBBAGCNwIBBDAcBgorBgEEAYI3AgELMQ4wDAYKKwYBBAGC
# NwIBFTAjBgkqhkiG9w0BCQQxFgQUK/bH7YuI+0geV383s94L45g8VmUwKgYKKwYB
# BAGCNwIBDDEcMBqhGIAWaHR0cDpcXHd3dy5hY3RpYW4uY29tIDANBgkqhkiG9w0B
# AQEFAASCAQBtY5wWP2TAoF0myaID6i46SLTuzPd9S9igf+6wXLcR4Mp7EpnGUtbE
# ZPz99pFJhmmli0babHybHMkDFsddPY6d+40EpimoKHFRCa3JFsauWX9JF82dO6fm
# qnh7FqUOFxAktMN16IbVLLB02SV2zSloWjpSPVgF6j9feQwY888x5NNTBWgP7ms2
# Wx33feqa9y6trfadfe9wYq1JS29piUL2JeLI6gx39RURAB2xKRHvPsMLCrrh2fLk
# 7El6J13YDHBg7dCDINDqK7f5hT8WDkZMKTq1a8PV24t/u9913LVaiuoUOBd0WQCH
# P1y/8aZ1uvp+vtbARal5XJ6n9QGbKTFMoYICCzCCAgcGCSqGSIb3DQEJBjGCAfgw
# ggH0AgEBMHIwXjELMAkGA1UEBhMCVVMxHTAbBgNVBAoTFFN5bWFudGVjIENvcnBv
# cmF0aW9uMTAwLgYDVQQDEydTeW1hbnRlYyBUaW1lIFN0YW1waW5nIFNlcnZpY2Vz
# IENBIC0gRzICEA7P9DjI/r81bgTYapgbGlAwCQYFKw4DAhoFAKBdMBgGCSqGSIb3
# DQEJAzELBgkqhkiG9w0BBwEwHAYJKoZIhvcNAQkFMQ8XDTE4MTAyMzE1MzA1N1ow
# IwYJKoZIhvcNAQkEMRYEFK/RfZGfWnbpWSRGAZLoIV2LzPREMA0GCSqGSIb3DQEB
# AQUABIIBAAw5YGknNAZFWOtv31MDQGFTvzDrAgGVTf5zXwGB3BoNwEeJ5jc4LIuL
# 0E5nEL7CtzuYX3nuLCSfpn05UxsctfBYX1dF9bwY4Ikt9F4GG6MyDQdQMouIUg8E
# kBMTkB9wFPb2XnQyYqFAa+BoSBFGP8+WJiaZ77ZTCDHpcGOYgBQHSgsHNeRZUq5/
# 3QPRCh+xP9Khqadhw94vtgFBipClSbi5JHQTYsvBYqNdB9qf8eNdJn/RKDusx4t+
# QSSE+H9wA9KT3X/NFZ554IVrGRve1k9Kji8AlXz9NjZiEO3o+Xt+YtvSw2T18hjc
# +c/JqU3bcC53z4RCnOIZHefcBsGsyAE=
# SIG # End signature block
